//===----------- PragmaOmpHandler.cpp - OpenMP pragma handling ------------===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file implements the class PragmaOmpHandler: the OpenMP
// pragma handler interface for the CodeGen compiler action.
//
//===----------------------------------------------------------------------===//

#include "clang/Lex/PragmaOmpHandler.h"
#include "clang/Lex/Preprocessor.h"
#include "clang/Lex/Token.h"
#include "clang/Lex/LexDiagnostic.h"
#include "llvm/Support/ErrorHandling.h"


using namespace clang;

//===----------------------------------------------------------------------===//
// PragmaOmpHandler construction.
//===----------------------------------------------------------------------===//
PragmaOmpHandler::PragmaOmpHandler() {
}

//===----------------------------------------------------------------------===//
// PragmaOmpHandler Destruction. 
//===----------------------------------------------------------------------===//
PragmaOmpHandler::~PragmaOmpHandler() {
  // Free all the OpenMP clause name objects which are constructed during the
  // initialization.
  for (int i=0; i<END_CLAUSE; i++)
  {
     delete Clause[i];
  }

  // Free all the default clause kind objects which are constructed during the
  // initialization.
  for (int i=0; i<D_END; i++)
  {
     delete DefaultClause[i];
  }

  // Free all the schedule clause kind objects which are constructed during the
  // initialization.
  for (int i=0; i<S_END; i++)
  {
     delete ScheduleClause[i];
  }
}

//===----------------------------------------------------------------------===//
// PragmaOmpHandler::Initialize
// \Brief: This function initializes all the data members of OpenMP pragma
//         handler interface.
//===----------------------------------------------------------------------===//
void PragmaOmpHandler::Initialize() {
  // We are just starting off with the translation unit. Initialize the flag
  // which tracks of, if the "unknown pragma warning message" is emitted for
  // this translation unit or not, when the user has not specified '-fopenmp'
  // flag in the command line.
  ResetPragmaOmpUnknownWarning();

  // Construct the "OpenMP Clause names" array. These names are referenced
  // while parsing OpenMP clauses.
  Clause[DEFAULT] = new StringRef("default");
  Clause[SHARED] = new StringRef("shared");
  Clause[PRIVATE] = new StringRef("private");
  Clause[FIRST_PRIVATE] = new StringRef("firstprivate");
  Clause[LAST_PRIVATE] = new StringRef("lastprivate");
  Clause[REDUCTION] = new StringRef("reduction");
  Clause[COPY_IN] = new StringRef("copyin");
  Clause[COPY_PRIVATE] = new StringRef("copyprivate");
  Clause[IF] = new StringRef("if");
  Clause[NUM_THREADS] = new StringRef("num_threads");
  Clause[FINAL] = new StringRef("final");
  Clause[SCHEDULE] = new StringRef("schedule");
  Clause[COLLAPSE] = new StringRef("collapse");
  Clause[ORDERED] = new StringRef("ordered");
  Clause[NOWAIT] = new StringRef("nowait");
  Clause[UNTIED] = new StringRef("untied");
  Clause[MERGEABLE] = new StringRef("mergeable");
  Clause[READ] = new StringRef("read");
  Clause[WRITE] = new StringRef("write");
  Clause[UPDATE] = new StringRef("update");
  Clause[CAPTURE] = new StringRef("capture");

  // Construct the "default clause kind" array. These names are referenced
  // while parsing default clause kind.
  DefaultClause[D_SHARED] = new StringRef("shared");
  DefaultClause[D_NONE] = new StringRef("none");

  // Construct the "schedule clause kind" array. These names are referenced
  // while parsing schedule clause kind.
  ScheduleClause[S_STATIC] = new StringRef("static"); 
  ScheduleClause[S_DYNAMIC] = new StringRef("dynamic");
  ScheduleClause[S_GUIDED] = new StringRef("guided");
  ScheduleClause[S_AUTO] = new StringRef("auto");
  ScheduleClause[S_RUNTIME] = new StringRef("runtime");

} // End of PragmaOmpHandler::Initialize()

//===----------------------------------------------------------------------===//
// Handle \#pragma omp parallel ...
//===----------------------------------------------------------------------===//
void PragmaOmpHandler::HandlePragmaOmpParallel(Preprocessor &PP,
                                               Token &OmpParaTok) {
  // \Brief: Check if the user has passed '-fopenmp' flag - If so, enter the 
  // *token* which is representing the current OpenMP pragma directive into
  // TokenStream so that parser can recognizes it and can parse the respective 
  // OpenMP pragma directive statement. Otherwise, report a warning that the
  // current OpenMP directive statement will be ignored.
  if (!PP.getLangOpts().OpenMpOn) {
    if (!GetPragmaOmpUnknownWarning()) {
      PP.Diag(OmpParaTok, diag::omp_pragma_ignored);
      SetPragmaOmpUnknownWarning();
    }
  } else {
    Token *Toks = new Token[1];
    Toks[0].startToken();
    Toks[0].setKind(tok::pragma_omp_parallel);
    Toks[0].setLocation(OmpParaTok.getLocation());
    PP.EnterTokenStream(Toks, 1, /*DisableMacroExpansion=*/true,
                        /*OwnsTokens=*/false);
  }
}

//===----------------------------------------------------------------------===//
// Handle \#pragma omp for ...
//===----------------------------------------------------------------------===//
void PragmaOmpHandler::HandlePragmaOmpFor(Preprocessor &PP,
                                          Token &OmpForTok) {
  // \Brief: Check if the user has passed '-fopenmp' flag - If so, enter the
  // *token* which is representing the current OpenMP pragma directive into
  // TokenStream so that parser can recognizes it and can parse the respective
  // OpenMP pragma directive statement. Otherwise, report a warning that the
  // current OpenMP directive statement will be ignored.
  if (!PP.getLangOpts().OpenMpOn) {
    if (!GetPragmaOmpUnknownWarning()) {
      PP.Diag(OmpForTok, diag::omp_pragma_ignored);
      SetPragmaOmpUnknownWarning();
    }
  } else {
    Token *Toks = new Token[1];
    Toks[0].startToken();
    Toks[0].setKind(tok::pragma_omp_for);
    Toks[0].setLocation(OmpForTok.getLocation());
    PP.EnterTokenStream(Toks, 1, /*DisableMacroExpansion=*/true,
                        /*OwnsTokens=*/false);
  }
}

//===----------------------------------------------------------------------===//
// Handle \#pragma omp sections ...
//===----------------------------------------------------------------------===//
void PragmaOmpHandler::HandlePragmaOmpSections(Preprocessor &PP,
                                               Token &OmpSectionsTok) {
  // \Brief: Check if the user has passed '-fopenmp' flag - If so, enter the
  // *token* which is representing the current OpenMP pragma directive into
  // TokenStream so that parser can recognizes it and can parse the respective
  // OpenMP pragma directive statement. Otherwise, report a warning that the
  // current OpenMP directive statement will be ignored.
  if (!PP.getLangOpts().OpenMpOn) {
    if (!GetPragmaOmpUnknownWarning()) {
      PP.Diag(OmpSectionsTok, diag::omp_pragma_ignored);
      SetPragmaOmpUnknownWarning();
    }
  } else {
    Token *Toks = new Token[1];
    Toks[0].startToken();
    Toks[0].setKind(tok::pragma_omp_sections);
    Toks[0].setLocation(OmpSectionsTok.getLocation());
    PP.EnterTokenStream(Toks, 1, /*DisableMacroExpansion=*/true,
                        /*OwnsTokens=*/false);
  }
}

//===----------------------------------------------------------------------===//
// Handle \#pragma omp section
//===----------------------------------------------------------------------===//
void PragmaOmpHandler::HandlePragmaOmpSection(Preprocessor &PP, 
                                              Token &OmpSectionTok) {
  // \Brief: Check if the user has passed '-fopenmp' flag - If so, enter the
  // *token* which is representing the current OpenMP pragma directive into
  // TokenStream so that parser can recognizes it and can parse the respective
  // OpenMP pragma directive statement. Otherwise, report a warning that the
  // current OpenMP directive statement will be ignored.
  if (!PP.getLangOpts().OpenMpOn) {
    if (!GetPragmaOmpUnknownWarning()) {
      PP.Diag(OmpSectionTok, diag::omp_pragma_ignored);
      SetPragmaOmpUnknownWarning();
    }
  } else {
    Token *Toks = new Token[1];
    Toks[0].startToken();
    Toks[0].setKind(tok::pragma_omp_section);
    Toks[0].setLocation(OmpSectionTok.getLocation());
    PP.EnterTokenStream(Toks, 1, /*DisableMacroExpansion=*/true,
                        /*OwnsTokens=*/false);
  }
}

//===----------------------------------------------------------------------===//
// Handle \#pragma omp single ...
//===----------------------------------------------------------------------===//
void PragmaOmpHandler::HandlePragmaOmpSingle(Preprocessor &PP, 
                                             Token &OmpSingleTok) {
  // \Brief: Check if the user has passed '-fopenmp' flag - If so, enter the
  // *token* which is representing the current OpenMP pragma directive into
  // TokenStream so that parser can recognizes it and can parse the respective
  // OpenMP pragma directive statement. Otherwise, report a warning that the
  // current OpenMP directive statement will be ignored.
  if (!PP.getLangOpts().OpenMpOn) {
    if (!GetPragmaOmpUnknownWarning()) {
      PP.Diag(OmpSingleTok, diag::omp_pragma_ignored);
      SetPragmaOmpUnknownWarning();
    }
  } else {
    Token *Toks = new Token[1];
    Toks[0].startToken();
    Toks[0].setKind(tok::pragma_omp_single);
    Toks[0].setLocation(OmpSingleTok.getLocation());
    PP.EnterTokenStream(Toks, 1, /*DisableMacroExpansion=*/true,
                        /*OwnsTokens=*/false);
  }
}

//===----------------------------------------------------------------------===//
// Handle \#pragma omp task ...
//===----------------------------------------------------------------------===//
void PragmaOmpHandler::HandlePragmaOmpTask(Preprocessor &PP,
                                           Token &OmpTaskTok) {
  // \Brief: Check if the user has passed '-fopenmp' flag - If so, enter the
  // *token* which is representing the current OpenMP pragma directive into
  // TokenStream so that parser can recognizes it and can parse the respective
  // OpenMP pragma directive statement. Otherwise, report a warning that the
  // current OpenMP directive statement will be ignored.
  if (!PP.getLangOpts().OpenMpOn) {
    if (!GetPragmaOmpUnknownWarning()) {
      PP.Diag(OmpTaskTok, diag::omp_pragma_ignored);
      SetPragmaOmpUnknownWarning();
    }
  } else {
    Token *Toks = new Token[1];
    Toks[0].startToken();
    Toks[0].setKind(tok::pragma_omp_task);
    Toks[0].setLocation(OmpTaskTok.getLocation());
    PP.EnterTokenStream(Toks, 1, /*DisableMacroExpansion=*/true,
                        /*OwnsTokens=*/false);
  }
}

//===----------------------------------------------------------------------===//
// Handle \#pragma omp master
//===----------------------------------------------------------------------===//
void PragmaOmpHandler::HandlePragmaOmpMaster(Preprocessor &PP,
                                             Token &OmpMasterTok) {
  // \Brief: Check if the user has passed '-fopenmp' flag - If so, enter the
  // *token* which is representing the current OpenMP pragma directive into
  // TokenStream so that parser can recognizes it and can parse the respective
  // OpenMP pragma directive statement. Otherwise, report a warning that the
  // current OpenMP directive statement will be ignored.
  if (!PP.getLangOpts().OpenMpOn) {
    if (!GetPragmaOmpUnknownWarning()) {
      PP.Diag(OmpMasterTok, diag::omp_pragma_ignored);
      SetPragmaOmpUnknownWarning();
    }
  } else {
    Token *Toks = new Token[1];
    Toks[0].startToken();
    Toks[0].setKind(tok::pragma_omp_master);
    Toks[0].setLocation(OmpMasterTok.getLocation());
    PP.EnterTokenStream(Toks, 1, /*DisableMacroExpansion=*/true,
                        /*OwnsTokens=*/false);
  }
}

//===----------------------------------------------------------------------===//
// Handle \#pragma omp critical ...
//===----------------------------------------------------------------------===//
void PragmaOmpHandler::HandlePragmaOmpCritical(Preprocessor &PP,
                                               Token &OmpCriticalTok) {
  // \Brief: Check if the user has passed '-fopenmp' flag - If so, enter the
  // *token* which is representing the current OpenMP pragma directive into
  // TokenStream so that parser can recognizes it and can parse the respective
  // OpenMP pragma directive statement. Otherwise, report a warning that the
  // current OpenMP directive statement will be ignored.
  if (!PP.getLangOpts().OpenMpOn) {
    if (!GetPragmaOmpUnknownWarning()) {
      PP.Diag(OmpCriticalTok, diag::omp_pragma_ignored);
      SetPragmaOmpUnknownWarning();
    }
  } else {
    Token *Toks = new Token[1];
    Toks[0].startToken();
    Toks[0].setKind(tok::pragma_omp_critical);
    Toks[0].setLocation(OmpCriticalTok.getLocation());
    PP.EnterTokenStream(Toks, 1, /*DisableMacroExpansion=*/true,
                        /*OwnsTokens=*/false);
  }
}

//===----------------------------------------------------------------------===//
// Handle \#pragma omp barrier
//===----------------------------------------------------------------------===//
void PragmaOmpHandler::HandlePragmaOmpBarrier(Preprocessor &PP, 
                                              Token &OmpBarrierTok) {
  // \Brief: Check if the user has passed '-fopenmp' flag - If so, enter the
  // *token* which is representing the current OpenMP pragma directive into
  // TokenStream so that parser can recognizes it and can parse the respective
  // OpenMP pragma directive statement. Otherwise, report a warning that the
  // current OpenMP directive statement will be ignored.
  if (!PP.getLangOpts().OpenMpOn) {
    if (!GetPragmaOmpUnknownWarning()) {
      PP.Diag(OmpBarrierTok, diag::omp_pragma_ignored);
      SetPragmaOmpUnknownWarning();
    }
  } else {
    Token *Toks = new Token[1];
    Toks[0].startToken();
    Toks[0].setKind(tok::pragma_omp_barrier);
    Toks[0].setLocation(OmpBarrierTok.getLocation());
    PP.EnterTokenStream(Toks, 1, /*DisableMacroExpansion=*/true,
                        /*OwnsTokens=*/false);
  }
}

//===----------------------------------------------------------------------===//
// Handle \#pragma omp taskwait
//===----------------------------------------------------------------------===//
void PragmaOmpHandler::HandlePragmaOmpTaskwait(Preprocessor &PP, 
                                               Token &OmpTaskwaitTok) {
  // \Brief: Check if the user has passed '-fopenmp' flag - If so, enter the
  // *token* which is representing the current OpenMP pragma directive into
  // TokenStream so that parser can recognizes it and can parse the respective
  // OpenMP pragma directive statement. Otherwise, report a warning that the
  // current OpenMP directive statement will be ignored.
  if (!PP.getLangOpts().OpenMpOn) {
    if (!GetPragmaOmpUnknownWarning()) {
      PP.Diag(OmpTaskwaitTok, diag::omp_pragma_ignored);
      SetPragmaOmpUnknownWarning();
    }
  } else {
    Token *Toks = new Token[1];
    Toks[0].startToken();
    Toks[0].setKind(tok::pragma_omp_taskwait);
    Toks[0].setLocation(OmpTaskwaitTok.getLocation());
    PP.EnterTokenStream(Toks, 1, /*DisableMacroExpansion=*/true,
                        /*OwnsTokens=*/false);
  }
}

//===----------------------------------------------------------------------===//
// Handle \#pragma omp taskyield
//===----------------------------------------------------------------------===//
void PragmaOmpHandler::HandlePragmaOmpTaskyield(Preprocessor &PP,
                                                Token &OmpTaskyieldTok) {
  // \Brief: Check if the user has passed '-fopenmp' flag - If so, enter the
  // *token* which is representing the current OpenMP pragma directive into
  // TokenStream so that parser can recognizes it and can parse the respective
  // OpenMP pragma directive statement. Otherwise, report a warning that the
  // current OpenMP directive statement will be ignored.
  if (!PP.getLangOpts().OpenMpOn) {
    if (!GetPragmaOmpUnknownWarning()) {
      PP.Diag(OmpTaskyieldTok, diag::omp_pragma_ignored);
      SetPragmaOmpUnknownWarning();
    }
  } else {
    Token *Toks = new Token[1];
    Toks[0].startToken();
    Toks[0].setKind(tok::pragma_omp_taskyield);
    Toks[0].setLocation(OmpTaskyieldTok.getLocation());
    PP.EnterTokenStream(Toks, 1, /*DisableMacroExpansion=*/true,
                        /*OwnsTokens=*/false);
  }
}

//===----------------------------------------------------------------------===//
// Handle \#pragma omp atomic
//===----------------------------------------------------------------------===//
void PragmaOmpHandler::HandlePragmaOmpAtomic(Preprocessor &PP, 
                                             Token &OmpAtomicTok) {
  // \Brief: Check if the user has passed '-fopenmp' flag - If so, enter the
  // *token* which is representing the current OpenMP pragma directive into
  // TokenStream so that parser can recognizes it and can parse the respective
  // OpenMP pragma directive statement. Otherwise, report a warning that the
  // current OpenMP directive statement will be ignored.
  if (!PP.getLangOpts().OpenMpOn) {
    if (!GetPragmaOmpUnknownWarning()) {
      PP.Diag(OmpAtomicTok, diag::omp_pragma_ignored);
      SetPragmaOmpUnknownWarning();
    }
  } else {
    Token *Toks = new Token[1];
    Toks[0].startToken();
    Toks[0].setKind(tok::pragma_omp_atomic);
    Toks[0].setLocation(OmpAtomicTok.getLocation());
    PP.EnterTokenStream(Toks, 1, /*DisableMacroExpansion=*/true,
                        /*OwnsTokens=*/false);
  }
}

//===----------------------------------------------------------------------===//
// Handle \#pragma omp flush ...
//===----------------------------------------------------------------------===//
void PragmaOmpHandler::HandlePragmaOmpFlush(Preprocessor &PP, 
                                            Token &OmpFlushTok) {
  // \Brief: Check if the user has passed '-fopenmp' flag - If so, enter the
  // *token* which is representing the current OpenMP pragma directive into
  // TokenStream so that parser can recognizes it and can parse the respective
  // OpenMP pragma directive statement. Otherwise, report a warning that the
  // current OpenMP directive statement will be ignored.
  if (!PP.getLangOpts().OpenMpOn) {
    if (!GetPragmaOmpUnknownWarning()) {
      PP.Diag(OmpFlushTok, diag::omp_pragma_ignored);
      SetPragmaOmpUnknownWarning();
    }
  } else {
    Token *Toks = new Token[1];
    Toks[0].startToken();
    Toks[0].setKind(tok::pragma_omp_flush);
    Toks[0].setLocation(OmpFlushTok.getLocation());
    PP.EnterTokenStream(Toks, 1, /*DisableMacroExpansion=*/true,
                        /*OwnsTokens=*/false);
  }
}

//===----------------------------------------------------------------------===//
// Handle \#pragma omp ordered
//===----------------------------------------------------------------------===//
void PragmaOmpHandler::HandlePragmaOmpOrdered(Preprocessor &PP,
                                              Token &OmpOrderedTok) {
  // \Brief: Check if the user has passed '-fopenmp' flag - If so, enter the
  // *token* which is representing the current OpenMP pragma directive into
  // TokenStream so that parser can recognizes it and can parse the respective
  // OpenMP pragma directive statement. Otherwise, report a warning that the
  // current OpenMP directive statement will be ignored.
  if (!PP.getLangOpts().OpenMpOn) {
    if (!GetPragmaOmpUnknownWarning()) {
      PP.Diag(OmpOrderedTok, diag::omp_pragma_ignored);
      SetPragmaOmpUnknownWarning();
    }
  } else {
    Token *Toks = new Token[1];
    Toks[0].startToken();
    Toks[0].setKind(tok::pragma_omp_ordered);
    Toks[0].setLocation(OmpOrderedTok.getLocation());
    PP.EnterTokenStream(Toks, 1, /*DisableMacroExpansion=*/true,
                        /*OwnsTokens=*/false);
  }
}

//===----------------------------------------------------------------------===//
// Handle \#pragma omp threadprivate ...
//===----------------------------------------------------------------------===//
void PragmaOmpHandler::HandlePragmaOmpThreadprivate(Preprocessor &PP, 
                                                    Token &OmpTpTok) {
  // \Brief: Check if the user has passed '-fopenmp' flag - If so, enter the
  // *token* which is representing the current OpenMP pragma directive into
  // TokenStream so that parser can recognizes it and can parse the respective
  // OpenMP pragma directive statement. Otherwise, report a warning that the
  // current OpenMP directive statement will be ignored.
  if (!PP.getLangOpts().OpenMpOn) {
    if (!GetPragmaOmpUnknownWarning()) {
      PP.Diag(OmpTpTok, diag::omp_pragma_ignored);
      SetPragmaOmpUnknownWarning();
    }
  } else {
    Token *Toks = new Token[1];
    Toks[0].startToken();
    Toks[0].setKind(tok::pragma_omp_threadprivate);
    Toks[0].setLocation(OmpTpTok.getLocation());
    PP.EnterTokenStream(Toks, 1, /*DisableMacroExpansion=*/true,
                        /*OwnsTokens=*/false);
  }
}

//===----------------------------------------------------------------------===//
// RegisterOmpPragmas - Install the different OpenMP preprocessor pragmas in
// the omp pragma namespace:
// \#pragma omp ...
//===----------------------------------------------------------------------===//
void PragmaOmpHandler::RegesterHandlersToPreprocessor(
                        Preprocessor &PP) {
  // \#pragma omp parallel ...
  PP.AddPragmaHandler("omp", new PragmaOmpParallelHandler(this));

   // \#pragma omp for ...
  PP.AddPragmaHandler("omp", new PragmaOmpForHandler(this));

  // \#pragma omp sections ...
  PP.AddPragmaHandler("omp", new PragmaOmpSectionsHandler(this));

  // \#pragma omp section
  PP.AddPragmaHandler("omp", new PragmaOmpSectionHandler(this));

  // \#pragma omp single
  PP.AddPragmaHandler("omp", new PragmaOmpSingleHandler(this));

  // \#pragma omp task ...
  PP.AddPragmaHandler("omp", new PragmaOmpTaskHandler(this));

  // \#pragma omp master
  PP.AddPragmaHandler("omp", new PragmaOmpMasterHandler(this));

  // \#pragma omp critical ...
  PP.AddPragmaHandler("omp", new PragmaOmpCriticalHandler(this));

  // \#pragma omp barrier
  PP.AddPragmaHandler("omp", new PragmaOmpBarrierHandler(this));

  // \#pragma omp taskwait
  PP.AddPragmaHandler("omp", new PragmaOmpTaskwaitHandler(this));

  // \#pragma omp taskyield
  PP.AddPragmaHandler("omp", new PragmaOmpTaskyieldHandler(this));

  // \#pragma omp atomic
  PP.AddPragmaHandler("omp", new PragmaOmpAtomicHandler(this));

  // \#pragma omp flush ...
  PP.AddPragmaHandler("omp", new PragmaOmpFlushHandler(this));

  // \#pragma omp ordered
  PP.AddPragmaHandler("omp", new PragmaOmpOrderedHandler(this));

  // \#pragma omp threadprivate ...
  PP.AddPragmaHandler("omp", new PragmaOmpThreadprivateHandler(this));
}
