//===--------- PragmaOmpHandler.h - OpenMP pragma handling ----------------===//
//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file defines the OpenMP pragma handler interface for CodeGen action.
//
//===----------------------------------------------------------------------===//

#ifndef LLVM_CLANG_PRAGMAOMPHANDLER_H
#define LLVM_CLANG_PRAGMAOMPHANDLER_H

#include "clang/Lex/Pragma.h"
#include "clang/Basic/LLVM.h"
#include "llvm/ADT/StringMap.h"
#include "llvm/ADT/StringRef.h"
#include <cassert>

namespace clang {

class Preprocessor;
class Token;
class SourceLocation;
class Parser;
class Sema;


//===----------------------------------------------------------------------===//
// PragmaOmpHandler - Instance of this interface is defined to handle the 
// various OpenMP pragmas when the compiler action to be executed is the 
// CodeGen action.
//===----------------------------------------------------------------------===//
class PragmaOmpHandler {
  // Make a Clang parser, the friend of OpenMP pragma handler, so that the
  // parser can have a access to the private members of the OpnMP pragma
  // handler.
  friend class Parser;

  // Make a Clang semantic analyser, the friend of OpenMP pragma handler, so
  // that the semantic analyser can have a access to the private members of the
  // OpenMP pragma handler.
  friend class Sema;

  // Enum values which locate different OpenMP clause names in the array of
  // OpenMP clause string objects.
  // Note: These enum values *must* be in consistant in *size* and *order* with
  // that of enum values defined in the AST node class "OmpClause".
  enum { 
    // OpenMP Clauses with a List of Variables.
    PRIVATE, FIRST_PRIVATE, LAST_PRIVATE, SHARED, COPY_IN, COPY_PRIVATE,
    // OpenMP Clauses with Scalar or Integer Expressions.
    IF, NUM_THREADS, FINAL, COLLAPSE, 
    // OpenMP Compound Clauses.
    REDUCTION, SCHEDULE, 
    // OpenMP Clauses with Predefined Values.
    DEFAULT, ORDERED, NOWAIT, UNTIED, MERGEABLE, READ, WRITE, UPDATE, CAPTURE,
    // End of enumarators.
    END_CLAUSE 
  };

  // Enum values which represents default clause *kind*.
  // Note: These enum values *must* be in consistant in *size* and *order* with
  // that of enum values defined in the AST node class "OmpDefaultClause".
  enum DefaultKind { D_SHARED, D_NONE, D_END };

  // Enum values which represents reduction clause operators *kind*.
  // Note: These enum values *must* be in consistant in *size* and *order* with
  // that of enum values defined in the AST node class "OmpReductionClause".
  //FIXME: Add remaining operators.
  enum ReductionOpKind { R_PLUS, R_MINUS, R_PROD, R_END };

  // Enum values which represents schedule clause *kind*.
  // Note: These enum values *must* be in consistant in *size* and *order* with
  // that of enum values defined in the AST node class "OmpScheduleClause".
  enum ScheduleKind { S_STATIC, S_DYNAMIC, S_GUIDED, S_AUTO, S_RUNTIME, S_END };

  // Private data members 
private:
  // \Brief: Holds all the OpenMP clause name strings. 
  // \Brief: These strings are referenced while parsing OpenMP clauses. 
  // Note that we won't introduce new *tokens* for openMP clause names
  // as these will get conflict with *identifier* token, and is very tricky
  // to handle. Instead, we reference below strings to recognize the OpenMP
  // clause names.
  StringRef* Clause[END_CLAUSE];

  // \Brief: Holds the strings for all the default clause kinds.
  // \Brief: These strings are referenced while parsing OpenMP clauses. 
  // Note that we won't introduce new *tokens* for default clause kinds
  // as these will get conflict with *identifier* token, and is very tricky
  // to handle. Instead, we reference below strings to recognize the default 
  // clause kind. 
  StringRef* DefaultClause[D_END];

  // \Brief: Holds the strings for all the schedule clause kinds.
  // \Brief: These strings are referenced while parsing OpenMP clauses. 
  // Note that we won't introduce new *tokens* for default clause kinds
  // as these will get conflict with *identifier* token, and is very tricky
  // to handle. Instead, we reference below strings to recognize the default 
  // clause kind. 
  StringRef* ScheduleClause[S_END];

  // PragmaOmpUnknownWarning - Holds the status of whether the unkown
  // pragma warning message is emitted for the current translation unit.
  // Note that clang emits only one warning message per translation unit
  // in order to avoid message bloating on the output device.
  bool PragmaOmpUnknownWarning;

  // Constructors and Destructors.
public:
  // Constructors
  PragmaOmpHandler(); 

  // Destructors.
  ~PragmaOmpHandler();

  // \Brief: Different OpenMP pragma handler routines. 
  // \Brief: These routines recieve calls from their respective hendler
  // interfaces, and initiate the handling of OpenMP constructs that they
  // are responsible for.
public:
  // \#pragma omp parallel ...
  void HandlePragmaOmpParallel(Preprocessor &PP, Token &OmpParaTok);
  // \#pragam omp for ...
  void HandlePragmaOmpFor(Preprocessor &PP, Token &OmpForTok);
  // \#pragma omp sections ...
  void HandlePragmaOmpSections(Preprocessor &PP, Token &OmpSectionsTok);
  // \#pragma omp section
  void HandlePragmaOmpSection(Preprocessor &PP, Token &OmpSectionTok);
  // \#pragma omp single ...
  void HandlePragmaOmpSingle(Preprocessor &PP, Token &OmpSingleTok);
  // \#pragma omp task ...
  void HandlePragmaOmpTask(Preprocessor &PP, Token &OmpTaskTok);
  // \#pragma omp master
  void HandlePragmaOmpMaster(Preprocessor &PP, Token &OmpMasterTok);
  // \#pragma omp critical ...
  void HandlePragmaOmpCritical(Preprocessor &PP, Token &OmpCriticalTok);
  // \#pragma omp barrier
  void HandlePragmaOmpBarrier(Preprocessor &PP, Token &OmpBarrierTok);
  // \#pragma omp taskwait
  void HandlePragmaOmpTaskwait(Preprocessor &PP, Token &OmpTaskwaitTok);
  // \#pragma omp taskyield
  void HandlePragmaOmpTaskyield(Preprocessor &PP, Token &OmpTaskyieldTok);
  // \#pragma omp atomic
  void HandlePragmaOmpAtomic(Preprocessor &PP, Token &OmpAtomicTok);
  // \#pragma omp flush ...
  void HandlePragmaOmpFlush(Preprocessor &PP, Token &OmpFlushTok);
  // \#pragma omp ordered
  void HandlePragmaOmpOrdered(Preprocessor &PP, Token &OmpOrderedTok);
  // \#pragma omp threadprivate ...
  void HandlePragmaOmpThreadprivate(Preprocessor &PP, Token &OmpTpTok);


  // \Brief: Different OpenMP pragma handler interfaces for the registration
  // of OpenMP pragma directives with Preprocessor.
private:
  /// PragmaOmpParallelHandler - "\#pragma omp parallel ...".
  struct PragmaOmpParallelHandler : public PragmaHandler {
    PragmaOmpHandler *Omp;
    PragmaOmpParallelHandler(PragmaOmpHandler *O)
	         : PragmaHandler("parallel"), Omp(O) {}
    virtual ~PragmaOmpParallelHandler() { Omp = 0; }
    virtual void HandlePragma(Preprocessor &PP, PragmaIntroducerKind Introducer,
			      Token &OmpParaTok) {
      Omp->HandlePragmaOmpParallel(PP, OmpParaTok);
    }
  }; // end of struct PragmaOmpParallelHandler

  /// PragmaOmpForHandler - "\#pragma omp for ...".
  struct PragmaOmpForHandler : public PragmaHandler {
    PragmaOmpHandler *Omp;
    PragmaOmpForHandler(PragmaOmpHandler *O)
            : PragmaHandler("for"), Omp(O) { }
    virtual ~PragmaOmpForHandler() { Omp = 0; }
    virtual void HandlePragma(Preprocessor &PP, PragmaIntroducerKind Introducer,
                            Token &OmpForTok) {
      Omp->HandlePragmaOmpFor(PP, OmpForTok);
    }
  }; // end of struct PragmaOmpForHandler

  /// PragmaOmpSectionsHandler - "\#pragma omp sections ...".
  struct PragmaOmpSectionsHandler : public PragmaHandler {
    PragmaOmpHandler *Omp;
    PragmaOmpSectionsHandler(PragmaOmpHandler *O)
            : PragmaHandler("sections"), Omp(O) { }
    virtual ~PragmaOmpSectionsHandler() { Omp = 0; }
    virtual void HandlePragma(Preprocessor &PP, PragmaIntroducerKind Introducer,
                            Token &OmpSectionsTok) {
      Omp->HandlePragmaOmpSections(PP, OmpSectionsTok);
    }
  }; // end for struct PragmaOmpSectionsHandler

  /// PragmaOmpSectionHandler - "\#pragma omp section".
  struct PragmaOmpSectionHandler : public PragmaHandler {
    PragmaOmpHandler *Omp;
    PragmaOmpSectionHandler(PragmaOmpHandler *O)
            : PragmaHandler("section"), Omp(O) { }
    virtual ~PragmaOmpSectionHandler() { Omp = 0; }
    virtual void HandlePragma(Preprocessor &PP, PragmaIntroducerKind Introducer,
                            Token &OmpSectionTok) {
      Omp->HandlePragmaOmpSection(PP, OmpSectionTok);
    }
  }; // end for struct PragmaOmpSectionHandler

  /// PragmaOmpSingleHandler - "\#pragma omp single ...".
  struct PragmaOmpSingleHandler : public PragmaHandler {
    PragmaOmpHandler *Omp;
    PragmaOmpSingleHandler(PragmaOmpHandler *O)
            : PragmaHandler("single"), Omp(O) { }
    virtual ~PragmaOmpSingleHandler() { Omp = 0; }
    virtual void HandlePragma(Preprocessor &PP, PragmaIntroducerKind Introducer,
                            Token &OmpSingleTok) {
      Omp->HandlePragmaOmpSingle(PP, OmpSingleTok);
    }
  }; // end for struct PragmaOmpSingleHandler

  /// PragmaOmpTaskHandler - "\#pragma omp task ...".
  struct PragmaOmpTaskHandler : public PragmaHandler {
    PragmaOmpHandler *Omp;
    PragmaOmpTaskHandler(PragmaOmpHandler *O)
            : PragmaHandler("task"), Omp(O) { }
    virtual ~PragmaOmpTaskHandler() { Omp = 0; }
    virtual void HandlePragma(Preprocessor &PP, PragmaIntroducerKind Introducer,
                            Token &OmpTaskTok) {
      Omp->HandlePragmaOmpTask(PP, OmpTaskTok);
    }
  }; // end for struct PragmaOmpTaskHandler

  /// PragmaOmpMasterHandler - "\#pragma omp master".
  struct PragmaOmpMasterHandler : public PragmaHandler {
    PragmaOmpHandler *Omp;
    PragmaOmpMasterHandler(PragmaOmpHandler *O)
            : PragmaHandler("master"), Omp(O) { }
    virtual ~PragmaOmpMasterHandler() { Omp = 0; }
    virtual void HandlePragma(Preprocessor &PP, PragmaIntroducerKind Introducer,
                            Token &OmpMasterTok) {
      Omp->HandlePragmaOmpMaster(PP, OmpMasterTok);
    }
  }; // end for struct PragmaOmpMasterHandler

  /// PragmaOmpCriticalHandler - "\#pragma omp critical ...".
  struct PragmaOmpCriticalHandler : public PragmaHandler {
    PragmaOmpHandler *Omp;
    PragmaOmpCriticalHandler(PragmaOmpHandler *O)
            : PragmaHandler("critical"), Omp(O) { }
    virtual ~PragmaOmpCriticalHandler() { Omp = 0; }
    virtual void HandlePragma(Preprocessor &PP, PragmaIntroducerKind Introducer,
                            Token &OmpCriticalTok) {
      Omp->HandlePragmaOmpCritical(PP, OmpCriticalTok);
    }
  }; // end for struct PragmaOmpMasterHandler

  /// PragmaOmpBarrierHandler - "\#pragma omp barrier".
  struct PragmaOmpBarrierHandler : public PragmaHandler {
    PragmaOmpHandler *Omp;
    PragmaOmpBarrierHandler(PragmaOmpHandler *O)
            : PragmaHandler("barrier"), Omp(O) { }
    virtual ~PragmaOmpBarrierHandler() { Omp = 0; }
    virtual void HandlePragma(Preprocessor &PP, PragmaIntroducerKind Introducer,
                            Token &OmpBarrierTok) {
      Omp->HandlePragmaOmpBarrier(PP, OmpBarrierTok);
    }
  }; // end for struct PragmaOmpBarrierHandler

  /// PragmaOmpTaskwaitHandler - "\#pragma omp taskwait".
  struct PragmaOmpTaskwaitHandler : public PragmaHandler {
    PragmaOmpHandler *Omp;
    PragmaOmpTaskwaitHandler(PragmaOmpHandler *O)
            : PragmaHandler("taskwait"), Omp(O) { }
    virtual ~PragmaOmpTaskwaitHandler() { Omp = 0; }
    virtual void HandlePragma(Preprocessor &PP, PragmaIntroducerKind Introducer,
                            Token &OmpTaskwaitTok) {
      Omp->HandlePragmaOmpTaskwait(PP, OmpTaskwaitTok);
    }
  }; // end for struct PragmaOmpTaskwaitHandler

  struct PragmaOmpTaskyieldHandler : public PragmaHandler {
    PragmaOmpHandler *Omp;
    PragmaOmpTaskyieldHandler(PragmaOmpHandler *O)
            : PragmaHandler("taskwait"), Omp(O) { }
    virtual ~PragmaOmpTaskyieldHandler() { Omp = 0; }
    virtual void HandlePragma(Preprocessor &PP, PragmaIntroducerKind Introducer,
                            Token &OmpTaskyieldTok) {
      Omp->HandlePragmaOmpTaskyield(PP, OmpTaskyieldTok);
    }
  }; // end for struct PragmaOmpTaskyieldHandler

  /// PragmaOmpAtomicHandler - "\#pragma omp atomic".
  struct PragmaOmpAtomicHandler : public PragmaHandler {
    PragmaOmpHandler *Omp;
    PragmaOmpAtomicHandler(PragmaOmpHandler *O)
            : PragmaHandler("atomic"), Omp(O) { }
    virtual ~PragmaOmpAtomicHandler() { Omp = 0; }
    virtual void HandlePragma(Preprocessor &PP, PragmaIntroducerKind Introducer,
                            Token &OmpAtomicTok) {
      Omp->HandlePragmaOmpAtomic(PP, OmpAtomicTok);
    }
  }; // end for struct PragmaOmpAtomicHandler 

  /// PragmaOmpFlushHandler - "\#pragma omp flush ...".
  struct PragmaOmpFlushHandler : public PragmaHandler {
    PragmaOmpHandler *Omp;
    PragmaOmpFlushHandler(PragmaOmpHandler *O)
            : PragmaHandler("flush"), Omp(O) { }
    virtual ~PragmaOmpFlushHandler() { }
    virtual void HandlePragma(Preprocessor &PP, PragmaIntroducerKind Introducer,
                            Token &OmpFlushTok) {
      Omp->HandlePragmaOmpFlush(PP, OmpFlushTok);
    }
  }; // end for struct PragmaOmpFlushHandler

  /// PragmaOmpOrderedHandler - "\#pragma omp ordered".
  struct PragmaOmpOrderedHandler : public PragmaHandler {
    PragmaOmpHandler *Omp;
    PragmaOmpOrderedHandler(PragmaOmpHandler *O)
            : PragmaHandler("ordered"), Omp(O)  { }
    virtual ~PragmaOmpOrderedHandler() { Omp = 0; }
    virtual void HandlePragma(Preprocessor &PP, PragmaIntroducerKind Introducer,
                            Token &OmpOrderedTok) {
      Omp->HandlePragmaOmpOrdered(PP, OmpOrderedTok);
    }
  }; // end for struct PragmaOmpOrderedHandler

  /// PragmaOmpThreadprivateHandler - "\#pragma omp threadprivate ...".
  struct PragmaOmpThreadprivateHandler : public PragmaHandler {
    PragmaOmpHandler *Omp;
    PragmaOmpThreadprivateHandler(PragmaOmpHandler *O)
            : PragmaHandler("threadprivate"), Omp(O) { }
    virtual ~PragmaOmpThreadprivateHandler() { Omp = 0; }
    virtual void HandlePragma(Preprocessor &PP, PragmaIntroducerKind Introducer,
                            Token &OmpTpTok) {
      Omp->HandlePragmaOmpThreadprivate(PP, OmpTpTok);
    }
  }; // end for struct PragmaOmpThreadprivateHandler

public:
  // When source contains omp pragmas, but user did not pass the 'fopenmp' flag,
  // then emit a warning message, but only once per source file.
  void SetPragmaOmpUnknownWarning() { PragmaOmpUnknownWarning = true; }
  void ResetPragmaOmpUnknownWarning() { PragmaOmpUnknownWarning = false; }
  bool GetPragmaOmpUnknownWarning() { return PragmaOmpUnknownWarning; }

  // Register OpenMP pragma handlers to Preprocessor so that Preprocessor
  // recognizes OpenMP pragmas.
  void RegesterHandlersToPreprocessor(Preprocessor &PP);

  // OmpPragmaHandler Initialization.
  void Initialize();

}; // end of class PragmaOmpHandler

}  // end namespace clang

#endif
