//===--------------------------- StmtOpenMP.cpp ---------------------------===//
//            /{ OpenMP directive statement AST Node Implementation }/
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file implements the OpenMP directive statement AST Node classes.
//
//===----------------------------------------------------------------------===//

#include "clang/AST/StmtOpenMP.h"
#include "clang/AST/Stmt.h"
#include "clang/AST/ExprCXX.h"
#include "clang/AST/ExprObjC.h"
#include "clang/AST/StmtCXX.h"
#include "clang/AST/StmtObjC.h"
#include "clang/AST/Type.h"
#include "clang/AST/ASTContext.h"
#include "clang/AST/ASTDiagnostic.h"
#include "clang/Basic/TargetInfo.h"
#include "llvm/ADT/StringExtras.h"
#include "llvm/Support/raw_ostream.h"
using namespace clang;

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node classes representing different OpenMP
// clauses. 
//===----------------------------------------------------------------------===//

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpClause" }/ which 
// represents the parent AST class node which holds all other omp /{ "clause" }
// AST nodes as its children.
//===----------------------------------------------------------------------===//
OmpClause::OmpClause(ASTContext &C, Stmt **Clauses[END_EXPR], 
                     unsigned NC[END_EXPR], SourceLocation SL,
                     SourceLocation EL)
  : Stmt(OmpClauseClass), ClauseStartLoc(SL), ClauseEndLoc(EL) {

  // Create and attach private clause list.
  if (NC[PRIVATE] > 0) {
    //FIXME: Is memory allocation and memcpy really required here?
    unsigned NumPrivateClauses = NC[PRIVATE]; 
    Stmt** PrivateClause = new (C) Stmt*[NumPrivateClauses];
    memcpy(PrivateClause, Clauses[PRIVATE], 
           NumPrivateClauses * sizeof(*PrivateClause));
    for (unsigned i=0; i<NumPrivateClauses-1; i++) {
      OmpPrivateClause* P = 
                          reinterpret_cast<OmpPrivateClause*>(PrivateClause[i]);
      P->NextOmpPrivateClause = PrivateClause[i+1];
    }
    SubExprs[PRIVATE] = PrivateClause[0];
  } else {
    SubExprs[PRIVATE] = 0;
  }

  // Create and attach firstprivate clause list.
  if (NC[FIRST_PRIVATE] > 0) {
    //FIXME: Is memory allocation and memcpy really required here?
    unsigned NumFirstPrivateClauses = NC[FIRST_PRIVATE];
    Stmt** FirstPrivateClause = new (C) Stmt*[NumFirstPrivateClauses];
    memcpy(FirstPrivateClause, Clauses[FIRST_PRIVATE],
           NumFirstPrivateClauses * sizeof(*FirstPrivateClause));
    for (unsigned i=0; i<NumFirstPrivateClauses-1; i++) {
      OmpFirstPrivateClause* FP =
            reinterpret_cast<OmpFirstPrivateClause*>(FirstPrivateClause[i]);
      FP->NextOmpFirstPrivateClause = FirstPrivateClause[i+1];
    }
    SubExprs[FIRST_PRIVATE] = FirstPrivateClause[0];
  } else {
    SubExprs[FIRST_PRIVATE] = 0;
  }

  // Create and attach lastprivate clause list.
  if (NC[LAST_PRIVATE] > 0) {
    //FIXME: Is memory allocation and memcpy really required here?
    unsigned NumLastPrivateClauses = NC[LAST_PRIVATE];
    Stmt** LastPrivateClause = new (C) Stmt*[NumLastPrivateClauses];
    memcpy(LastPrivateClause, Clauses[LAST_PRIVATE],
           NumLastPrivateClauses * sizeof(*LastPrivateClause));
    for (unsigned i=0; i<NumLastPrivateClauses-1; i++) {
      OmpLastPrivateClause* LP =
             reinterpret_cast<OmpLastPrivateClause*>(LastPrivateClause[i]);
      LP->NextOmpLastPrivateClause = LastPrivateClause[i+1];
    }
    SubExprs[LAST_PRIVATE] = LastPrivateClause[0];
  } else {
    SubExprs[LAST_PRIVATE] = 0;
  }

  // Create and attach shared clause list.
  if (NC[SHARED] > 0) {
    //FIXME: Is memory allocation and memcpy really required here?
    unsigned NumSharedClauses = NC[SHARED];
    Stmt** SharedClause = new (C) Stmt*[NumSharedClauses];
    memcpy(SharedClause, Clauses[SHARED],
           NumSharedClauses * sizeof(*SharedClause));
    for (unsigned i=0; i<NumSharedClauses-1; i++) {
      OmpSharedClause* SP =
                          reinterpret_cast<OmpSharedClause*>(SharedClause[i]);
      SP->NextOmpSharedClause = SharedClause[i+1];
    }
    SubExprs[COPY_IN] = SharedClause[0];
  } else {
    SubExprs[COPY_IN] = 0;
  }

  // Create and attach copyin clause list.
  if (NC[COPY_IN] > 0) {
    //FIXME: Is memory allocation and memcpy really required here?
    unsigned NumCopyInClauses = NC[COPY_IN];
    Stmt** CopyInClause = new (C) Stmt*[NumCopyInClauses];
    memcpy(CopyInClause, Clauses[COPY_IN],
           NumCopyInClauses * sizeof(*CopyInClause));
    for (unsigned i=0; i<NumCopyInClauses-1; i++) {
      OmpCopyInClause* CP =
                          reinterpret_cast<OmpCopyInClause*>(CopyInClause[i]);
      CP->NextOmpCopyInClause = CopyInClause[i+1];
    }
    SubExprs[COPY_IN] = CopyInClause[0];
  } else {
    SubExprs[COPY_IN] = 0;
  }

  // Create and attach copyprivate clause list.
  if (NC[COPY_PRIVATE] > 0) {
    //FIXME: Is memory allocation and memcpy really required here?
    unsigned NumCopyPrivateClauses = NC[COPY_PRIVATE];
    Stmt** CopyPrivateClause = new (C) Stmt*[NumCopyPrivateClauses];
    memcpy(CopyPrivateClause, Clauses[COPY_PRIVATE],
           NumCopyPrivateClauses * sizeof(*CopyPrivateClause));
    for (unsigned i=0; i<NumCopyPrivateClauses-1; i++) {
      OmpCopyPrivateClause* CP =
               reinterpret_cast<OmpCopyPrivateClause*>(CopyPrivateClause[i]);
      CP->NextOmpCopyPrivateClause = CopyPrivateClause[i+1];
    }
    SubExprs[COPY_PRIVATE] = CopyPrivateClause[0];
  } else {
    SubExprs[COPY_PRIVATE] = 0;
  }

  // Create and attach if clause list.
  if (NC[IF] > 0) {
    //FIXME: Assert that vector size is exactly one.
    //FIXME: Is memory allocation and memcpy really required here?
    Stmt** IfClause = new (C) Stmt*[1];
    memcpy(IfClause, Clauses[IF], sizeof(*IfClause));
    SubExprs[IF] = IfClause[0];
  } else {
    SubExprs[IF] = 0;
  }

  // Create and attach numthreads clause list.
  if (NC[NUM_THREADS] > 0) {
    //FIXME: Assert that vector size is exactly one.
    //FIXME: Is memory allocation and memcpy really required here?
    Stmt** NumThreadsClause = new (C) Stmt*[1];
    memcpy(NumThreadsClause, Clauses[NUM_THREADS], sizeof(*NumThreadsClause));
    SubExprs[NUM_THREADS] = NumThreadsClause[0];
  } else {
    SubExprs[NUM_THREADS] = 0;
  }

  // Create and attach final clause list.
  if (NC[FINAL] > 0) {
    //FIXME: Assert that vector size is exactly one.
    //FIXME: Is memory allocation and memcpy really required here?
    Stmt** FinalClause = new (C) Stmt*[1];
    memcpy(FinalClause, Clauses[FINAL], sizeof(*FinalClause));
    SubExprs[FINAL] = FinalClause[0];
  } else {
    SubExprs[FINAL] = 0;
  }

  // Create and attach collapse clause list.
  if (NC[COLLAPSE] > 0) {
    //FIXME: Assert that vector size is exactly one.
    //FIXME: Is memory allocation and memcpy really required here?
    Stmt** CollapseClause = new (C) Stmt*[1];
    memcpy(CollapseClause, Clauses[COLLAPSE], sizeof(*CollapseClause));
    SubExprs[COLLAPSE] = CollapseClause[0];
  } else {
    SubExprs[COLLAPSE] = 0;
  }

  // Create and attach reduction clause list.
  if (NC[REDUCTION] > 0) {
    //FIXME: Is memory allocation and memcpy really required here?
    unsigned NumReductionClauses = NC[REDUCTION];
    Stmt** ReductionClause = new (C) Stmt*[NumReductionClauses];
    memcpy(ReductionClause, Clauses[REDUCTION],
           NumReductionClauses * sizeof(*ReductionClause));
    for (unsigned i=0; i<NumReductionClauses-1; i++) {
      OmpReductionClause* R =
           reinterpret_cast<OmpReductionClause*>(ReductionClause[i]);
      R->NextOmpReductionClause = ReductionClause[i+1];
    }
    SubExprs[REDUCTION] = ReductionClause[0];
  } else {
    SubExprs[REDUCTION] = 0;
  }

  // Create and attach schedule clause list.
  if (NC[SCHEDULE] > 0) {
    //FIXME: Assert that vector size is exactly one.
    //FIXME: Is memory allocation and memcpy really required here?
    Stmt** ScheduleClause = new (C) Stmt*[1];
    memcpy(ScheduleClause, Clauses[SCHEDULE], sizeof(*ScheduleClause));
    SubExprs[SCHEDULE] = ScheduleClause[0];
  } else {
    SubExprs[SCHEDULE] = 0;
  }

  // Create and attach default clause list.
  if (NC[DEFAULT] > 0) {
    //FIXME: Assert that vector size is exactly one.
    //FIXME: Is memory allocation and memcpy really required here?
    Stmt** DefaultClause = new (C) Stmt*[1];
    memcpy(DefaultClause, Clauses[DEFAULT], sizeof(*DefaultClause));
    SubExprs[DEFAULT] = DefaultClause[0];
  } else {
    SubExprs[DEFAULT] = 0;
  }

  // Create and attach ordered clause list.
  if (NC[ORDERED] > 0) {
    //FIXME: Assert that vector size is exactly one.
    //FIXME: Is memory allocation and memcpy really required here?
    Stmt** OrderedClause = new (C) Stmt*[1];
    memcpy(OrderedClause, Clauses[ORDERED], sizeof(*OrderedClause));
    SubExprs[ORDERED] = OrderedClause[0];
  } else {
    SubExprs[ORDERED] = 0;
  }

  // Create and attach nowait clause list.
  if (NC[NOWAIT] > 0) {
    //FIXME: Assert that vector size is exactly one.
    //FIXME: Is memory allocation and memcpy really required here?
    Stmt** NoWaitClause = new (C) Stmt*[1];
    memcpy(NoWaitClause, Clauses[NOWAIT], sizeof(*NoWaitClause));
    SubExprs[NOWAIT] = NoWaitClause[0];
  } else {
    SubExprs[NOWAIT] = 0;
  }

  // Create and attach untied clause list.
  if (NC[UNTIED] > 0) {
    //FIXME: Assert that vector size is exactly one.
    //FIXME: Is memory allocation and memcpy really required here?
    Stmt** UntiedClause = new (C) Stmt*[1];
    memcpy(UntiedClause, Clauses[UNTIED], sizeof(*UntiedClause));
    SubExprs[UNTIED] = UntiedClause[0];
  } else {
    SubExprs[UNTIED] = 0;
  }

  // Create and attach mergeable clause list.
  if (NC[MERGEABLE] > 0) {
    //FIXME: Assert that vector size is exactly one.
    //FIXME: Is memory allocation and memcpy really required here?
    Stmt** MergeableClause = new (C) Stmt*[1];
    memcpy(MergeableClause, Clauses[MERGEABLE], sizeof(*MergeableClause));
    SubExprs[MERGEABLE] = MergeableClause[0];
  } else {
    SubExprs[MERGEABLE] = 0;
  }

  // Create and attach read clause list.
  if (NC[READ] > 0) {
    //FIXME: Assert that vector size is exactly one.
    //FIXME: Is memory allocation and memcpy really required here?
    Stmt** ReadClause = new (C) Stmt*[1];
    memcpy(ReadClause, Clauses[READ], sizeof(*ReadClause));
    SubExprs[READ] = ReadClause[0];
  } else {
    SubExprs[READ] = 0;
  }

  // Create and attach write clause list.
  if (NC[WRITE] > 0) {
    //FIXME: Assert that vector size is exactly one.
    //FIXME: Is memory allocation and memcpy really required here?
    Stmt** WriteClause = new (C) Stmt*[1];
    memcpy(WriteClause, Clauses[WRITE], sizeof(*WriteClause));
    SubExprs[WRITE] = WriteClause[0];
  } else {
    SubExprs[WRITE] = 0;
  }

  // Create and attach update clause list.
  if (NC[UPDATE] > 0) {
    //FIXME: Assert that vector size is exactly one.
    //FIXME: Is memory allocation and memcpy really required here?
    Stmt** UpdateClause = new (C) Stmt*[1];
    memcpy(UpdateClause, Clauses[UPDATE], sizeof(*UpdateClause));
    SubExprs[UPDATE] = UpdateClause[0];
  } else {
    SubExprs[UPDATE] = 0;
  }

  // Create and attach capture clause list.
  if (NC[CAPTURE] > 0) {
    //FIXME: Assert that vector size is exactly one.
    //FIXME: Is memory allocation and memcpy really required here?
    Stmt** CaptureClause = new (C) Stmt*[1];
    memcpy(CaptureClause, Clauses[CAPTURE], sizeof(*CaptureClause));
    SubExprs[CAPTURE] = CaptureClause[0];
  } else {
    SubExprs[CAPTURE] = 0;
  }
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpPrivateClause" }/ which
// represents the OpenMP pragma clause  /{ "private". }/.
//===----------------------------------------------------------------------===//
OmpPrivateClause::OmpPrivateClause(ASTContext &C, Expr **ExprStart,
                                   unsigned NumExprs, SourceLocation PC,
                                   SourceLocation LP, SourceLocation RP)
  : Stmt(OmpPrivateClauseClass), PrivateLoc(PC), LParenLoc(LP),
    RParenLoc(RP) {
  OmpClauseBits.NumDeclRefExprs = NumExprs; 
  assert(OmpClauseBits.NumDeclRefExprs == NumExprs &&
         "NumExprs doesn't fit in bits of OmpClauseBits.NumDeclRefExprs!");

  DeclRefExprs = new (C) Stmt*[NumExprs];
  memcpy(DeclRefExprs, ExprStart, NumExprs * sizeof(*DeclRefExprs));
  NextOmpPrivateClause = 0;
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpFirstPrivateClause" }/
// which represents the OpenMP pragma clause  /{ "firstprivate". }/.
//===----------------------------------------------------------------------===//
OmpFirstPrivateClause::OmpFirstPrivateClause(ASTContext &C, Expr **ExprStart,
                                             unsigned NumExprs,
                                             SourceLocation RPC,
                                             SourceLocation LP,
                                             SourceLocation RP)
  : Stmt(OmpFirstPrivateClauseClass), FirstPrivateLoc(RPC), LParenLoc(LP),
    RParenLoc(RP) {
  OmpClauseBits.NumDeclRefExprs = NumExprs;
  assert(OmpClauseBits.NumDeclRefExprs == NumExprs &&
         "NumExprs doesn't fit in bits of OmpClauseBits.NumDeclRefExprs!");

  DeclRefExprs = new (C) Stmt*[NumExprs];
  memcpy(DeclRefExprs, ExprStart, NumExprs * sizeof(*DeclRefExprs));
  NextOmpFirstPrivateClause = 0;
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpLastPrivateClause" }/ 
// which represents the OpenMP pragma clause  /{ "lastprivate". }/.
//===----------------------------------------------------------------------===//
OmpLastPrivateClause::OmpLastPrivateClause(ASTContext &C, Expr **ExprStart,
                                           unsigned NumExprs,
                                           SourceLocation LPC,
                                           SourceLocation LP,
                                           SourceLocation RP)
  : Stmt(OmpLastPrivateClauseClass), LastPrivateLoc(LPC), LParenLoc(LP),
    RParenLoc(RP) {
  OmpClauseBits.NumDeclRefExprs = NumExprs;
  assert(OmpClauseBits.NumDeclRefExprs == NumExprs &&
         "NumExprs doesn't fit in bits of OmpClauseBits.NumDeclRefExprs!");

  DeclRefExprs = new (C) Stmt*[NumExprs];
  memcpy(DeclRefExprs, ExprStart, NumExprs * sizeof(*DeclRefExprs));
  NextOmpLastPrivateClause = 0;
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpSharedClause" }/ which
// represents the OpenMP pragma clause  /{ "shared". }/.
//===----------------------------------------------------------------------===//
OmpSharedClause::OmpSharedClause(ASTContext &C, Expr **ExprStart,
                                 unsigned NumExprs, SourceLocation SC,
                                 SourceLocation LP, SourceLocation RP)
  : Stmt(OmpSharedClauseClass), SharedLoc(SC), LParenLoc(LP),
    RParenLoc(RP) {
  OmpClauseBits.NumDeclRefExprs = NumExprs;
  assert(OmpClauseBits.NumDeclRefExprs == NumExprs &&
         "NumExprs doesn't fit in bits of OmpClauseBits.NumDeclRefExprs!");

  DeclRefExprs = new (C) Stmt*[NumExprs];
  memcpy(DeclRefExprs, ExprStart, NumExprs * sizeof(*DeclRefExprs));
  NextOmpSharedClause = 0;
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpCopyInClause" }/ which
// represents the OpenMP pragma clause  /{ "copyin". }/.
//===----------------------------------------------------------------------===//
OmpCopyInClause::OmpCopyInClause(ASTContext &C, Expr **ExprStart,
                                 unsigned NumExprs, SourceLocation CIC,
                                 SourceLocation LP, SourceLocation RP)
  : Stmt(OmpCopyInClauseClass), CopyInLoc(CIC), LParenLoc(LP),
    RParenLoc(RP) {
  OmpClauseBits.NumDeclRefExprs = NumExprs;
  assert(OmpClauseBits.NumDeclRefExprs == NumExprs &&
         "NumExprs doesn't fit in bits of OmpClauseBits.NumDeclRefExprs!");

  DeclRefExprs = new (C) Stmt*[NumExprs];
  memcpy(DeclRefExprs, ExprStart, NumExprs * sizeof(*DeclRefExprs));
  NextOmpCopyInClause = 0;
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpCopyPrivateClause" }/
// which represents the OpenMP pragma clause  /{ "copyprivate". }/.
//===----------------------------------------------------------------------===//
OmpCopyPrivateClause::OmpCopyPrivateClause(ASTContext &C, Expr **ExprStart,
                                           unsigned NumExprs, 
                                           SourceLocation CPC,
                                           SourceLocation LP, SourceLocation RP)
  : Stmt(OmpCopyPrivateClauseClass), CopyPrivateLoc(CPC), LParenLoc(LP),
    RParenLoc(RP) {
  OmpClauseBits.NumDeclRefExprs = NumExprs;
  assert(OmpClauseBits.NumDeclRefExprs == NumExprs &&
         "NumExprs doesn't fit in bits of OmpClauseBits.NumDeclRefExprs!");

  DeclRefExprs = new (C) Stmt*[NumExprs];
  memcpy(DeclRefExprs, ExprStart, NumExprs * sizeof(*DeclRefExprs));
  NextOmpCopyPrivateClause = 0;
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpIfClause" }/ which
// represents the OpenMP pragma clause  /{ "if". }/.
//===-------------/--------------------------------------------------------===//
OmpIfClause::OmpIfClause(ASTContext &C, SourceLocation IL, SourceLocation LPL,
                         SourceLocation RPL, VarDecl *var, Expr *cond)
  : Stmt(OmpIfClauseClass), IfLoc(IL), LParenLoc(LPL), RParenLoc(RPL) {
  setConditionVariable(C, var);
  SubExprs[COND] = reinterpret_cast<Stmt*>(cond);
}

void OmpIfClause::setConditionVariable(ASTContext &C, VarDecl *V) {
  if (!V) {
    SubExprs[VAR] = 0;
    return;
  }

  SourceRange VarRange = V->getSourceRange();
  SubExprs[VAR] = new (C) DeclStmt(DeclGroupRef(V), VarRange.getBegin(),
                                   VarRange.getEnd());
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpNumThreadsClause" }/
// which represents the OpenMP pragma clause  /{ "num_threads". }/.
//===----------------------------------------------------------------------===//
OmpNumThreadsClause::OmpNumThreadsClause(ASTContext &C, SourceLocation NTL,
                                         SourceLocation LPL, SourceLocation RPL,
                                         Expr *IExpr) 
  : Stmt(OmpNumThreadsClauseClass), NumThreadsLoc(NTL), LParenLoc(LPL),
    RParenLoc(RPL) {
  IntExpr = reinterpret_cast<Stmt*>(IExpr);
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpFinalClause" }/ which
// represents the OpenMP pragma clause  /{ "final". }/.
//===-------------/--------------------------------------------------------===//
OmpFinalClause::OmpFinalClause(ASTContext &C, SourceLocation FL, 
                               SourceLocation LPL, SourceLocation RPL,
                               VarDecl *var, Expr *cond)
  : Stmt(OmpFinalClauseClass), FinalLoc(FL), LParenLoc(LPL), RParenLoc(RPL) {
  setConditionVariable(C, var);
  SubExprs[COND] = reinterpret_cast<Stmt*>(cond);
} 
 
void OmpFinalClause::setConditionVariable(ASTContext &C, VarDecl *V) {
  if (!V) { 
    SubExprs[VAR] = 0;
    return; 
  } 
 
  SourceRange VarRange = V->getSourceRange();
  SubExprs[VAR] = new (C) DeclStmt(DeclGroupRef(V), VarRange.getBegin(),
                                   VarRange.getEnd());
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpCollapseClause" }/
// which represents the OpenMP pragma clause  /{ "collapse". }/.
//===----------------------------------------------------------------------===// 
OmpCollapseClause::OmpCollapseClause(ASTContext &C, SourceLocation CL, 
                                     SourceLocation LPL, SourceLocation RPL, 
                                     Expr *IExpr)   
  : Stmt(OmpCollapseClauseClass), CollapseLoc(CL), LParenLoc(LPL), 
    RParenLoc(RPL) {
  IntExpr = reinterpret_cast<Stmt*>(IExpr);
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpReductionClause" }/
// which represents the OpenMP pragma clause  /{ "reduction". }/.
//===----------------------------------------------------------------------===//
OmpReductionClause::OmpReductionClause(ASTContext &C, Expr **ExprStart,
                                       unsigned NumExprs, unsigned Kind,
                                       SourceLocation RC, SourceLocation LP,
                                       SourceLocation OP, SourceLocation CL,
                                       SourceLocation RP)
  : Stmt(OmpReductionClauseClass), ReductionLoc(RC), LParenLoc(LP), 
    OperatorLoc(OP), ColonLoc(CL), RParenLoc(RP) {
  OmpClauseBits.NumDeclRefExprs = NumExprs;
  assert(OmpClauseBits.NumDeclRefExprs == NumExprs &&
         "NumExprs doesn't fit in bits of OmpClauseBits.NumDeclRefExprs!");

  DeclRefExprs = new (C) Stmt*[NumExprs];
  memcpy(DeclRefExprs, ExprStart, NumExprs * sizeof(*DeclRefExprs));
  NextOmpReductionClause = 0;
  OpKind = (enum OperatorKind)Kind; 
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpScheduleClause" }/
// which represents the OpenMP pragma clause  /{ "schedule". }/.
//===----------------------------------------------------------------------===//
OmpScheduleClause::OmpScheduleClause(ASTContext &C, Expr *SExpr, unsigned Kind,
                                     SourceLocation SC, SourceLocation LP,
                                     SourceLocation SKL, SourceLocation RP)
  : Stmt(OmpScheduleClauseClass), ScheduleLoc(SC), LParenLoc(LP),
    ScheduleKindLoc(SKL), RParenLoc(RP) {
  SKind = (ScheduleKind)Kind;
  ScheduleExpr = reinterpret_cast<Stmt*>(SExpr);
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpDefaultClause" }/ which
// represents the OpenMP pragma clause  /{ "default". }/.
//===----------------------------------------------------------------------===//
OmpDefaultClause::OmpDefaultClause(ASTContext &C, unsigned K,
                                   SourceLocation DL, SourceLocation LPL,
                                   SourceLocation DKL, SourceLocation RPL)
  : Stmt(OmpDefaultClauseClass), DefaultLoc(DL), LParenLoc(LPL),
    DefaultKindLoc(DKL), RParenLoc(RPL) {
  Kind = (enum DefaultKind)K;
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpOrderedClause" }/ which
// represents the OpenMP pragma clause  /{ "ordered". }/.
//===----------------------------------------------------------------------===//
OmpOrderedClause::OmpOrderedClause(ASTContext &C, SourceLocation OL)
  : Stmt(OmpOrderedClauseClass), OrderedLoc(OL) {
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpNoWaitClause" }/ which
// represents the OpenMP pragma clause  /{ "nowait". }/.
//===----------------------------------------------------------------------===//
OmpNoWaitClause::OmpNoWaitClause(ASTContext &C, SourceLocation NWL) 
  : Stmt(OmpNoWaitClauseClass), NoWaitLoc(NWL) { 
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpUntiedClause" }/ which
// represents the OpenMP pragma clause  /{ "untied". }/.
//===----------------------------------------------------------------------===//
OmpUntiedClause::OmpUntiedClause(ASTContext &C, SourceLocation UL) 
  : Stmt(OmpUntiedClauseClass), UntiedLoc(UL) { 
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpMergeableClause" }/
// which represents the OpenMP pragma clause  /{ "mergeable". }/.
//===----------------------------------------------------------------------===//
OmpMergeableClause::OmpMergeableClause(ASTContext &C, SourceLocation ML) 
  : Stmt(OmpMergeableClauseClass), MergeableLoc(ML) { 
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpReadClause" }/
// which represents the OpenMP pragma clause  /{ "read". }/.
//===----------------------------------------------------------------------===//
OmpReadClause::OmpReadClause(ASTContext &C, SourceLocation RL)
  : Stmt(OmpReadClauseClass), ReadLoc(RL) {
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpWriteClause" }/
// which represents the OpenMP pragma clause  /{ "write". }/.
//===----------------------------------------------------------------------===//
OmpWriteClause::OmpWriteClause(ASTContext &C, SourceLocation WL)
  : Stmt(OmpWriteClauseClass), WriteLoc(WL) {
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpUpdateClause" }/
// which represents the OpenMP pragma clause  /{ "update". }/.
//===----------------------------------------------------------------------===//
OmpUpdateClause::OmpUpdateClause(ASTContext &C, SourceLocation UL)
  : Stmt(OmpUpdateClauseClass), UpdateLoc(UL) {
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpCaptureClause" }/
// which represents the OpenMP pragma clause  /{ "capture". }/.
//===----------------------------------------------------------------------===//
OmpCaptureClause::OmpCaptureClause(ASTContext &C, SourceLocation CL)
  : Stmt(OmpCaptureClauseClass), CaptureLoc(CL) {
} 

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node classes representing different OpenMP
// directive statements. 
//===----------------------------------------------------------------------===//

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpParallelStmt" }/ which
// represents the OpenMP pragma directive statement
// /{ \#pragma omp parallel ... }/. 
//===----------------------------------------------------------------------===//
OmpParallelStmt::OmpParallelStmt(ASTContext &C, Stmt *Clause,
                                 Stmt *ParallelBody, SourceLocation PL)
  : Stmt(OmpParallelStmtClass), ParallelLoc(PL) {
  SubExprs[CLAUSE] = Clause;
  SubExprs[PARALLEL_BODY] = ParallelBody;
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpForStmt" }/ which
// represents the OpenMP pragma directive statement
// /{ \#pragma omp for ... }/. 
//===----------------------------------------------------------------------===//
OmpForStmt::OmpForStmt(ASTContext &C, Stmt *Clause, Stmt *Body,
                       SourceLocation FL)
  : Stmt(OmpForStmtClass), ForLoc(FL) {
  SubExprs[CLAUSE] = Clause;
  SubExprs[BODY] = Body;
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpParallelForStmt" }/
// which represents the OpenMP pragma directive statement
// /{ \#pragma omp parallel for ... }/. 
//===----------------------------------------------------------------------===//
OmpParallelForStmt::OmpParallelForStmt(ASTContext &C, Stmt *Clause, Stmt *Body, 
                                       SourceLocation PFL)
  : Stmt(OmpParallelForStmtClass), ParallelForLoc(PFL) {
  SubExprs[CLAUSE] = Clause;
  SubExprs[BODY] = Body;
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpSectionsStmt" }/ which
// represents the OpenMP pragma directive statement
// /{ \#pragma omp sections ... }/. 
//===----------------------------------------------------------------------===//
OmpSectionsStmt::OmpSectionsStmt(ASTContext &C, Stmt *Clause, Stmt *Body, 
                                 SourceLocation SL)
  : Stmt(OmpSectionsStmtClass), SectionsLoc(SL) {
  SubExprs[CLAUSE] = Clause;
  SubExprs[BODY] = Body;
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpSectionStmt" }/ which
// represents the OpenMP pragma directive statement
// /{ \#pragma omp section }/. 
//===----------------------------------------------------------------------===//
OmpSectionStmt::OmpSectionStmt(ASTContext &C, Stmt *Body, SourceLocation SL)
  : Stmt(OmpSectionStmtClass), SectionLoc(SL) {
  SubStmt = Body;
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpSingleStmt" }/ which
// represents the OpenMP pragma directive statement
// /{ \#pragma omp single ... }/. 
//===----------------------------------------------------------------------===//
OmpSingleStmt::OmpSingleStmt(ASTContext &C, Stmt *Clause, Stmt *Body,
                             SourceLocation SL)
  : Stmt(OmpSingleStmtClass), SingleLoc(SL) {
  SubExprs[CLAUSE] = Clause;
  SubExprs[BODY] = Body;
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpTaskStmt" }/ which
// represents the OpenMP pragma directive statement
// /{ \#pragma omp task ... }/. 
//===----------------------------------------------------------------------===//
OmpTaskStmt::OmpTaskStmt(ASTContext &C, Stmt *Clause, Stmt *Body,
                         SourceLocation TL)
  : Stmt(OmpTaskStmtClass), TaskLoc(TL) {
  SubExprs[CLAUSE] = Clause;
  SubExprs[BODY] = Body;
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpMasterStmt" }/ which
// represents the OpenMP pragma directive statement
// /{ \#pragma omp master }/. 
//===----------------------------------------------------------------------===//
OmpMasterStmt::OmpMasterStmt(ASTContext &C, Stmt *Body, SourceLocation ML)
  : Stmt(OmpMasterStmtClass), MasterLoc(ML) {
  SubStmt = Body;
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpBarrierStmt" }/ which
// represents the OpenMP pragma directive statement
// /{ \#pragma omp barrier }/. 
//===----------------------------------------------------------------------===//
OmpBarrierStmt::OmpBarrierStmt(ASTContext &C, SourceLocation BL)
  : Stmt(OmpBarrierStmtClass), BarrierLoc(BL) {
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpTaskwaitStmt" }/ which
// represents the OpenMP pragma directive statement
// /{ \#pragma omp taskwait }/. 
//===----------------------------------------------------------------------===//
OmpTaskwaitStmt::OmpTaskwaitStmt(ASTContext &C, SourceLocation TL)
  : Stmt(OmpTaskwaitStmtClass), TaskwaitLoc(TL) {
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpTaskyieldStmt" }/ which
// represents the OpenMP pragma directive statement
// /{ \#pragma omp taskyield }/. 
//===----------------------------------------------------------------------===//
OmpTaskyieldStmt::OmpTaskyieldStmt(ASTContext &C, SourceLocation TL)
  : Stmt(OmpTaskyieldStmtClass), TaskyieldLoc(TL) {
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpOrderedStmt" }/ which
// represents the OpenMP pragma directive statement
// /{ \#pragma omp ordered }/. 
//===----------------------------------------------------------------------===//
OmpOrderedStmt::OmpOrderedStmt(ASTContext &C, Stmt *Body, SourceLocation OL)
  : Stmt(OmpOrderedStmtClass), OrderedLoc(OL) {
  SubStmt = Body;
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpThreadPrivateStmt" }/
// which represents the OpenMP pragma directive statement
// /{ \#pragma omp threadpribate... }/. 
//===----------------------------------------------------------------------===//
OmpThreadPrivateStmt::OmpThreadPrivateStmt(ASTContext &C, Expr **ExprStart, 
                                           unsigned NumExprs,
                                           SourceLocation TPC,
                                           SourceLocation LP,
                                           SourceLocation RP)
  : Stmt(OmpThreadPrivateStmtClass), ThreadPrivateLoc(TPC), LParenLoc(LP),
    RParenLoc(RP) {
  OmpClauseBits.NumDeclRefExprs = NumExprs;
  assert(OmpClauseBits.NumDeclRefExprs == NumExprs &&
         "NumExprs doesn't fit in bits of OmpClauseBits.NumDeclRefExprs!");

  DeclRefExprs = new (C) Stmt*[NumExprs];
  memcpy(DeclRefExprs, ExprStart, NumExprs * sizeof(*DeclRefExprs));
} 

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpFlushStmt" }/
// which represents the OpenMP pragma directive statement
// /{ \#pragma omp flush... }/. 
//===----------------------------------------------------------------------===//
OmpFlushStmt::OmpFlushStmt(ASTContext &C, Expr **ExprStart, unsigned NumExprs,
                           SourceLocation FC, SourceLocation LP,
                           SourceLocation RP)
  : Stmt(OmpFlushStmtClass), FlushLoc(FC), LParenLoc(LP), RParenLoc(RP) {
  OmpClauseBits.NumDeclRefExprs = NumExprs;
  assert(OmpClauseBits.NumDeclRefExprs == NumExprs &&
         "NumExprs doesn't fit in bits of OmpClauseBits.NumDeclRefExprs!");

  DeclRefExprs = new (C) Stmt*[NumExprs];
  memcpy(DeclRefExprs, ExprStart, NumExprs * sizeof(*DeclRefExprs)); 
}

//===----------------------------------------------------------------------===//
// \Brief: Implementation of AST node class /{ "class OmpAtomicStmt" }/
// which represents the OpenMP pragma directive statement
// /{ \#pragma omp atomic... }/. 
//===----------------------------------------------------------------------===//
OmpAtomicStmt::OmpAtomicStmt(ASTContext &C, Stmt *Clause, Stmt *Body,
                             SourceLocation AL, bool BodyIsExprStmt)
  : Stmt(OmpAtomicStmtClass), AtomicLoc(AL) {
  SubExprs[CLAUSE] = Clause;
  SubExprs[BODY] = Body;
  OmpClauseBits.AtomicStmtIsExpr = BodyIsExprStmt;
} 

