//===-- StmtOpenMP.h - Classes representing OpenMP directive statements ---===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file defines the OpenMP directive statement and OpenMP clause AST node
// classes.
//
//===----------------------------------------------------------------------===//

#ifndef LLVM_CLANG_AST_STMTOPENMP_H
#define LLVM_CLANG_AST_STMTOPENMP_H

#include "clang/AST/Stmt.h"
#include "llvm/Support/Compiler.h"

namespace clang {


//===----------------------------------------------------------------------===//
// AST node classes representing different OpenMP clauses 
//===----------------------------------------------------------------------===//

//===----------------------------------------------------------------------===//
// OmpClause - Parent AST class node which holds all other omp /{ "clause" }/
// AST nodes as its children.
//===----------------------------------------------------------------------===//
class OmpClause : public Stmt {
  // Note: These enum values *must* be in consistant in *size* and *order* with
  // that of enum values defined in the class "PragmaOmpHandler".
  enum { 
    // OpenMP Clauses with a List of Variables.
    PRIVATE, FIRST_PRIVATE, LAST_PRIVATE, SHARED, COPY_IN, COPY_PRIVATE,
    // OpenMP Clauses with Scalar or Integer Expressions.
    IF, NUM_THREADS, FINAL, COLLAPSE,      
    // OpenMP Compound Clauses.
    REDUCTION, SCHEDULE,     
    // OpenMP Clauses with Predefined Values.
    DEFAULT, ORDERED, NOWAIT, UNTIED, MERGEABLE, READ, WRITE, UPDATE, CAPTURE,
    // End of enumarators.
    END_EXPR 
  };
  Stmt* SubExprs[END_EXPR];
  SourceLocation ClauseStartLoc;
  SourceLocation ClauseEndLoc;

public:
  OmpClause(ASTContext &C, Stmt **Clauses[END_EXPR], unsigned NC[END_EXPR],
        SourceLocation SL, SourceLocation EL);

  explicit OmpClause(EmptyShell Empty)
                          : Stmt(OmpClauseClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    //FIXME
    return SourceRange(ClauseStartLoc, ClauseEndLoc);
  }

  child_range children() {
    //FIXME
    return child_range(&SubExprs[0], &SubExprs[0]+END_EXPR);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpClauseClass;
  }

  static bool classof(const OmpClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpPrivateClause - This class represents AST node for OpenMP pragma
// clause "private". 
//===----------------------------------------------------------------------===//
class OmpPrivateClause : public Stmt {
  friend class OmpClause;

  Stmt** DeclRefExprs;
  Stmt* NextOmpPrivateClause;
  SourceLocation PrivateLoc;
  SourceLocation LParenLoc;
  SourceLocation RParenLoc;

public:
  OmpPrivateClause(ASTContext &C, Expr **ExprStart, unsigned NumExprs,
        SourceLocation PC, SourceLocation LP, SourceLocation RP);

  explicit OmpPrivateClause(EmptyShell Empty)
                          : Stmt(OmpPrivateClauseClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(PrivateLoc, RParenLoc);
  }

  child_range children() {
    return child_range(&DeclRefExprs[0],
                       &DeclRefExprs[0]+OmpClauseBits.NumDeclRefExprs);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpPrivateClauseClass;
  }

  static bool classof(const OmpPrivateClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpFirstPrivateClause - This class represents AST node for OpenMP pragma
// clause "firstfirtprivate". 
//===----------------------------------------------------------------------===//
class OmpFirstPrivateClause : public Stmt {
  friend class OmpClause;

  Stmt** DeclRefExprs;
  Stmt* NextOmpFirstPrivateClause;
  SourceLocation FirstPrivateLoc;
  SourceLocation LParenLoc;
  SourceLocation RParenLoc;

public:
  OmpFirstPrivateClause(ASTContext &C, Expr **ExprStart, unsigned NumExprs,
        SourceLocation FPC, SourceLocation LP, SourceLocation RP);

  explicit OmpFirstPrivateClause(EmptyShell Empty)
                          : Stmt(OmpFirstPrivateClauseClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(FirstPrivateLoc, RParenLoc);
  }

  child_range children() {
    return child_range(&DeclRefExprs[0],
                       &DeclRefExprs[0]+OmpClauseBits.NumDeclRefExprs);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpFirstPrivateClauseClass;
  }

  static bool classof(const OmpFirstPrivateClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpLastPrivateClause - This class represents AST node for OpenMP pragma
// clause "lastprivate". 
//===----------------------------------------------------------------------===//
class OmpLastPrivateClause : public Stmt {
  friend class OmpClause;

  Stmt** DeclRefExprs;
  Stmt* NextOmpLastPrivateClause;
  SourceLocation LastPrivateLoc;
  SourceLocation LParenLoc;
  SourceLocation RParenLoc;

public:
  OmpLastPrivateClause(ASTContext &C, Expr **ExprStart, unsigned NumExprs,
        SourceLocation LPC, SourceLocation LP, SourceLocation RP);

  explicit OmpLastPrivateClause(EmptyShell Empty)
                          : Stmt(OmpLastPrivateClauseClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(LastPrivateLoc, RParenLoc);
  }

  child_range children() {
    return child_range(&DeclRefExprs[0],
                       &DeclRefExprs[0]+OmpClauseBits.NumDeclRefExprs);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpLastPrivateClauseClass;
  }

  static bool classof(const OmpLastPrivateClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpSharedClause - This class represents AST node for OpenMP pragma clause 
// "shared". 
//===----------------------------------------------------------------------===//
class OmpSharedClause : public Stmt {
  friend class OmpClause;

  Stmt** DeclRefExprs;
  Stmt* NextOmpSharedClause;
  SourceLocation SharedLoc;
  SourceLocation LParenLoc;
  SourceLocation RParenLoc;

public:
  OmpSharedClause(ASTContext &C, Expr **ExprStart, unsigned NumExprs,
        SourceLocation SC, SourceLocation LP, SourceLocation RP);

  explicit OmpSharedClause(EmptyShell Empty)
                          : Stmt(OmpSharedClauseClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(SharedLoc, RParenLoc);
  }

  child_range children() {
    return child_range(&DeclRefExprs[0],
                       &DeclRefExprs[0]+OmpClauseBits.NumDeclRefExprs);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpSharedClauseClass;
  }

  static bool classof(const OmpSharedClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpCopyInClause - This class represents AST node for OpenMP pragma clause 
// "copyin". 
//===----------------------------------------------------------------------===//
class OmpCopyInClause : public Stmt {
  friend class OmpClause;

  Stmt** DeclRefExprs;
  Stmt* NextOmpCopyInClause;
  SourceLocation CopyInLoc;
  SourceLocation LParenLoc;
  SourceLocation RParenLoc;

public:
  OmpCopyInClause(ASTContext &C, Expr **ExprStart, unsigned NumExprs,
        SourceLocation CIC, SourceLocation LP, SourceLocation RP);

  explicit OmpCopyInClause(EmptyShell Empty)
                          : Stmt(OmpCopyInClauseClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(CopyInLoc, RParenLoc);
  }

  child_range children() {
    return child_range(&DeclRefExprs[0],
                       &DeclRefExprs[0]+OmpClauseBits.NumDeclRefExprs);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpCopyInClauseClass;
  }

  static bool classof(const OmpCopyInClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpCopyPrivateClause - This class represents AST node for OpenMP pragma
// clause "copyprivate".
//===----------------------------------------------------------------------===//
class OmpCopyPrivateClause : public Stmt {
  friend class OmpClause;

  Stmt** DeclRefExprs;
  Stmt* NextOmpCopyPrivateClause;
  SourceLocation CopyPrivateLoc;
  SourceLocation LParenLoc;
  SourceLocation RParenLoc;

public:
  OmpCopyPrivateClause(ASTContext &C, Expr **ExprStart, unsigned NumExprs,
        SourceLocation CPC, SourceLocation LP, SourceLocation RP);

  explicit OmpCopyPrivateClause(EmptyShell Empty)
                          : Stmt(OmpCopyPrivateClauseClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(CopyPrivateLoc, RParenLoc);
  }

  child_range children() {
    return child_range(&DeclRefExprs[0],
                       &DeclRefExprs[0]+OmpClauseBits.NumDeclRefExprs);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpCopyPrivateClauseClass;
  }

  static bool classof(const OmpCopyPrivateClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpIfClause - This class represents AST node for OpenMP pragma clause "if". 
//===----------------------------------------------------------------------===//
class OmpIfClause : public Stmt {
  enum { VAR, COND, END_EXPR };
  Stmt* SubExprs[END_EXPR];

  SourceLocation IfLoc;
  SourceLocation LParenLoc;
  SourceLocation RParenLoc;

public:
  OmpIfClause(ASTContext &C, SourceLocation IL, SourceLocation LPL,
         SourceLocation RPL, VarDecl *var, Expr *cond);
  explicit OmpIfClause(EmptyShell Empty)
                          : Stmt(OmpIfClauseClass, Empty) { }

  void setConditionVariable(ASTContext &C, VarDecl *V);

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(IfLoc, RParenLoc);
  }

  child_range children() { 
     return child_range(&SubExprs[0], &SubExprs[0]+END_EXPR);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpIfClauseClass;
  }

  static bool classof(const OmpIfClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpNumThreadsClause - This class represents AST node for OpenMP pragma 
// clause "num_threads".
//===----------------------------------------------------------------------===//
class OmpNumThreadsClause : public Stmt {
  Stmt* IntExpr;

  SourceLocation NumThreadsLoc;
  SourceLocation LParenLoc;
  SourceLocation RParenLoc;

public:
  OmpNumThreadsClause(ASTContext &C, SourceLocation NTL, SourceLocation LPL,
         SourceLocation RPL, Expr *IntExpr);
  explicit OmpNumThreadsClause(EmptyShell Empty)
                          : Stmt(OmpNumThreadsClauseClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(NumThreadsLoc, RParenLoc);
  }

  child_range children() { return child_range(&IntExpr, &IntExpr+1); }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpNumThreadsClauseClass;
  }

  static bool classof(const OmpNumThreadsClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpFinalClause - This class represents AST node for OpenMP pragma clause
// "final". 
//===----------------------------------------------------------------------===//
class OmpFinalClause : public Stmt {
  enum { VAR, COND, END_EXPR };
  Stmt* SubExprs[END_EXPR];

  SourceLocation FinalLoc;
  SourceLocation LParenLoc;
  SourceLocation RParenLoc;

public:
  OmpFinalClause(ASTContext &C, SourceLocation FL, SourceLocation LPL,
         SourceLocation RPL, VarDecl *var, Expr *cond);
  explicit OmpFinalClause(EmptyShell Empty)
                          : Stmt(OmpFinalClauseClass, Empty) { }

  void setConditionVariable(ASTContext &C, VarDecl *V);

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(FinalLoc, RParenLoc);
  }

  child_range children() { 
     return child_range(&SubExprs[0], &SubExprs[0]+END_EXPR);
  } 

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpFinalClauseClass;
  }

  static bool classof(const OmpFinalClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpCollapseClause - This class represents AST node for OpenMP pragma 
// clause "collapse".
//===----------------------------------------------------------------------===//
class OmpCollapseClause : public Stmt {
  Stmt* IntExpr;

  SourceLocation CollapseLoc;
  SourceLocation LParenLoc;
  SourceLocation RParenLoc;

public:
  OmpCollapseClause(ASTContext &C, SourceLocation CL, SourceLocation LPL,
         SourceLocation RPL, Expr *IntExpr);
  explicit OmpCollapseClause(EmptyShell Empty)
                          : Stmt(OmpCollapseClauseClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(CollapseLoc, RParenLoc);
  } 

  child_range children() { return child_range(&IntExpr, &IntExpr+1); }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpCollapseClauseClass;
  }

  static bool classof(const OmpCollapseClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpReductionClause - This class represents AST node for OpenMP pragma
// clause "reduction".
//===----------------------------------------------------------------------===//
class OmpReductionClause : public Stmt {
  friend class OmpClause;
  
  // Note: These enum values *must* be in consistant in *size* and *order* with
  // that of enum values defined in the class "PragmaOmpHandler".
  //FIXME: Add remaining operators.
  enum OperatorKind { R_PLUS, R_MINUS, R_PROD, R_END };

  Stmt** DeclRefExprs;
  Stmt* NextOmpReductionClause;
  enum OperatorKind OpKind;
  SourceLocation ReductionLoc;
  SourceLocation LParenLoc;
  SourceLocation OperatorLoc;
  SourceLocation ColonLoc;
  SourceLocation RParenLoc;

public:
  OmpReductionClause(ASTContext &C, Expr **ExprStart, unsigned NumExprs,
        unsigned Kind, SourceLocation RC, SourceLocation LP, SourceLocation OP,
        SourceLocation CL, SourceLocation RP);

  explicit OmpReductionClause(EmptyShell Empty)
                          : Stmt(OmpReductionClauseClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(ReductionLoc, RParenLoc);
  }

  child_range children() {
    return child_range(&DeclRefExprs[0],
                       &DeclRefExprs[0]+OmpClauseBits.NumDeclRefExprs);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpReductionClauseClass;
  }

  static bool classof(const OmpReductionClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpScheduleClause - This class represents AST node for OpenMP pragma
// clause "schedule".
//===----------------------------------------------------------------------===//
class OmpScheduleClause : public Stmt {
  // Note: These enum values *must* be in consistant in *size* and *order* with
  // that of enum values defined in the class "PragmaOmpHandler".
  enum ScheduleKind { S_STATIC, S_DYNAMIC, S_GUIDED, S_AUTO, S_RUNTIME, S_END };

  Stmt* ScheduleExpr;
  enum ScheduleKind SKind;
  SourceLocation ScheduleLoc;
  SourceLocation LParenLoc;
  SourceLocation ScheduleKindLoc;
  SourceLocation RParenLoc;

public:
  OmpScheduleClause(ASTContext &C, Expr *SExpr, unsigned Kind, SourceLocation SC,
        SourceLocation LP, SourceLocation SKL, SourceLocation RP);

  explicit OmpScheduleClause(EmptyShell Empty)
                          : Stmt(OmpScheduleClauseClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(ScheduleLoc, RParenLoc);
  }

  child_range children() {
    return child_range(&ScheduleExpr, &ScheduleExpr+1);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpScheduleClauseClass;
  }

  static bool classof(const OmpScheduleClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpDefaultClause - This class represents AST node for OpenMP pragma
// clause "default". 
//===----------------------------------------------------------------------===//
class OmpDefaultClause : public Stmt {
  // Note: These enum values *must* be in consistant in *size* and *order* with
  // that of enum values defined in the class "PragmaOmpHandler".
  enum DefaultKind { D_SHARED, D_NONE, D_END };
  enum DefaultKind Kind;
  SourceLocation DefaultLoc;
  SourceLocation LParenLoc;
  SourceLocation DefaultKindLoc;
  SourceLocation RParenLoc;

public:
  OmpDefaultClause(ASTContext &C, unsigned Kind, SourceLocation DL,
                   SourceLocation LPL, SourceLocation DKL, SourceLocation RPL);
  explicit OmpDefaultClause(EmptyShell Empty)
                          : Stmt(OmpDefaultClauseClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(DefaultLoc, RParenLoc);
  }

  child_range children() { return child_range(); }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpDefaultClauseClass;
  }

  static bool classof(const OmpDefaultClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpOrderedClause - This class represents AST node for OpenMP pragma
// clause "ordered".
//===----------------------------------------------------------------------===//
class OmpOrderedClause : public Stmt {
  SourceLocation OrderedLoc;

public:
  OmpOrderedClause(ASTContext &C, SourceLocation OL);
  explicit OmpOrderedClause(EmptyShell Empty)
                          : Stmt(OmpOrderedClauseClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(OrderedLoc);
  }

  child_range children() { return child_range(); }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpOrderedClauseClass;
  }

  static bool classof(const OmpOrderedClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpNoWaitClause - This class represents AST node for OpenMP pragma
// clause "nowait".
//===----------------------------------------------------------------------===//
class OmpNoWaitClause : public Stmt {
  SourceLocation NoWaitLoc;

public:
  OmpNoWaitClause(ASTContext &C, SourceLocation NWL);
  explicit OmpNoWaitClause(EmptyShell Empty)
                          : Stmt(OmpNoWaitClauseClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(NoWaitLoc);
  }

  child_range children() { return child_range(); }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpNoWaitClauseClass;
  }

  static bool classof(const OmpNoWaitClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpUntiedClause - This class represents AST node for OpenMP pragma
// clause "untied".
//===----------------------------------------------------------------------===//
class OmpUntiedClause : public Stmt {
  SourceLocation UntiedLoc;

public:
  OmpUntiedClause(ASTContext &C, SourceLocation UL);
  explicit OmpUntiedClause(EmptyShell Empty)
                          : Stmt(OmpUntiedClauseClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(UntiedLoc);
  }

  child_range children() { return child_range(); }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpUntiedClauseClass;
  }

  static bool classof(const OmpUntiedClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpMergeableClause - This class represents AST node for OpenMP pragma
// clause "mergeable".
//===----------------------------------------------------------------------===//
class OmpMergeableClause : public Stmt {
  SourceLocation MergeableLoc;

public:
  OmpMergeableClause(ASTContext &C, SourceLocation ML);
  explicit OmpMergeableClause(EmptyShell Empty)
                          : Stmt(OmpMergeableClauseClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(MergeableLoc);
  }

  child_range children() { return child_range(); }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpMergeableClauseClass;
  }

  static bool classof(const OmpMergeableClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpReadClause - This class represents AST node for OpenMP pragma
// clause "read".
//===----------------------------------------------------------------------===//
class OmpReadClause : public Stmt {
  SourceLocation ReadLoc;

public:
  OmpReadClause(ASTContext &C, SourceLocation RL);
  explicit OmpReadClause(EmptyShell Empty)
                          : Stmt(OmpReadClauseClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(ReadLoc);
  }

  child_range children() { return child_range(); }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpReadClauseClass;
  }

  static bool classof(const OmpReadClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpWriteClause - This class represents AST node for OpenMP pragma
// clause "write".
//===----------------------------------------------------------------------===//
class OmpWriteClause : public Stmt {
  SourceLocation WriteLoc;
 
public:
  OmpWriteClause(ASTContext &C, SourceLocation WL);
  explicit OmpWriteClause(EmptyShell Empty)
                          : Stmt(OmpWriteClauseClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(WriteLoc);
  }

  child_range children() { return child_range(); }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpWriteClauseClass;
  }

  static bool classof(const OmpWriteClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpUpdateClause - This class represents AST node for OpenMP pragma
// clause "update".
//===----------------------------------------------------------------------===//
class OmpUpdateClause : public Stmt {
  SourceLocation UpdateLoc;
  
public:
  OmpUpdateClause(ASTContext &C, SourceLocation UL);
  explicit OmpUpdateClause(EmptyShell Empty)
                          : Stmt(OmpUpdateClauseClass, Empty) { }
 
  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(UpdateLoc);
  }
 
  child_range children() { return child_range(); }
 
  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpUpdateClauseClass;
  }
 
  static bool classof(const OmpUpdateClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpCaptureClause - This class represents AST node for OpenMP pragma
// clause "capture".
//===----------------------------------------------------------------------===//
class OmpCaptureClause : public Stmt {
  SourceLocation CaptureLoc;

public:
  OmpCaptureClause(ASTContext &C, SourceLocation CL);
  explicit OmpCaptureClause(EmptyShell Empty)
                          : Stmt(OmpCaptureClauseClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(CaptureLoc);
  }
 
  child_range children() { return child_range(); }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpCaptureClauseClass;
  }

  static bool classof(const OmpCaptureClause *) { return true; }
};

//===----------------------------------------------------------------------===//
// AST node classes representing different OpenMP directives 
//===----------------------------------------------------------------------===//

//===----------------------------------------------------------------------===//
// OmpParallelStmt - This class represents AST node for OpenMP directive 
// statement /{ "parallel" }/.
//===----------------------------------------------------------------------===//
class OmpParallelStmt : public Stmt {
  enum { CLAUSE, PARALLEL_BODY, END_EXPR };
  Stmt* SubExprs[END_EXPR];
  SourceLocation ParallelLoc;

public:
  OmpParallelStmt(ASTContext &C, Stmt *Clause, Stmt *ParallelBody,
                  SourceLocation PL);
  explicit OmpParallelStmt(EmptyShell Empty) 
                          : Stmt(OmpParallelStmtClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    //FIXME?
    return SourceRange(ParallelLoc, SubExprs[PARALLEL_BODY]->getLocEnd());
  }

  child_range children() {
    //FIXME?
    return child_range(&SubExprs[0], &SubExprs[0]+END_EXPR);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpParallelStmtClass;
  }

  static bool classof(const OmpParallelStmt *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpForStmt - This class represents AST node for OpenMP directive 
// statement /{ "for" }/.
//===----------------------------------------------------------------------===//
class OmpForStmt : public Stmt {
  enum { CLAUSE, BODY, END_EXPR };
  Stmt* SubExprs[END_EXPR];
  SourceLocation ForLoc;

public:
  OmpForStmt(ASTContext &C, Stmt *Clause, Stmt *Body, SourceLocation FL);
  explicit OmpForStmt(EmptyShell Empty)
                          : Stmt(OmpForStmtClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    //FIXME?
    return SourceRange(ForLoc, SubExprs[BODY]->getLocEnd());
  }

  child_range children() {
    //FIXME?
    return child_range(&SubExprs[0], &SubExprs[0]+END_EXPR);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpForStmtClass;
  }

  static bool classof(const OmpForStmt *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpParallelForStmt - This class represents AST node for OpenMP directive 
// statement /{ "parallel for" }/.
//===----------------------------------------------------------------------===//
class OmpParallelForStmt : public Stmt {
  enum { CLAUSE, BODY, END_EXPR };
  Stmt* SubExprs[END_EXPR];
  SourceLocation ParallelForLoc;

public:
  OmpParallelForStmt(ASTContext &C, Stmt *Clause, Stmt *Body,
                     SourceLocation PFL);
  explicit OmpParallelForStmt(EmptyShell Empty)
                          : Stmt(OmpParallelForStmtClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    //FIXME?
    return SourceRange(ParallelForLoc, SubExprs[BODY]->getLocEnd());
  }

  child_range children() {
    //FIXME?
    return child_range(&SubExprs[0], &SubExprs[0]+END_EXPR);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpParallelForStmtClass;
  }

  static bool classof(const OmpParallelForStmt *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpSectionsStmt - This class represents AST node for OpenMP directive 
// statement /{ "sections" }/.
//===----------------------------------------------------------------------===//
class OmpSectionsStmt : public Stmt {
  enum { CLAUSE, BODY, END_EXPR };
  Stmt* SubExprs[END_EXPR];
  SourceLocation SectionsLoc;

public:
  OmpSectionsStmt(ASTContext &C, Stmt *Clause, Stmt *Body, SourceLocation SL);
  explicit OmpSectionsStmt(EmptyShell Empty)
                          : Stmt(OmpSectionsStmtClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    //FIXME?
    return SourceRange(SectionsLoc, SubExprs[BODY]->getLocEnd());
  }

  child_range children() {
    //FIXME?
    return child_range(&SubExprs[0], &SubExprs[0]+END_EXPR);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpSectionsStmtClass;
  }

  static bool classof(const OmpSectionsStmt *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpSectionStmt - This class represents AST node for OpenMP directive 
// statement /{ "section" }/.
//===----------------------------------------------------------------------===//
class OmpSectionStmt : public Stmt {
  Stmt* SubStmt;
  SourceLocation SectionLoc;

public:
  OmpSectionStmt(ASTContext &C, Stmt *Body, SourceLocation SL);
  explicit OmpSectionStmt(EmptyShell Empty)
                          : Stmt(OmpSectionStmtClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    //FIXME?
    return SourceRange(SectionLoc, SubStmt->getLocEnd());
  }

  child_range children() {
    //FIXME?
    return child_range(&SubStmt, &SubStmt+1);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpSectionStmtClass;
  }

  static bool classof(const OmpSectionStmt *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpSingleStmt - This class represents AST node for OpenMP directive 
// statement /{ "single" }/.
//===----------------------------------------------------------------------===//
class OmpSingleStmt : public Stmt {
  enum { CLAUSE, BODY, END_EXPR };
  Stmt* SubExprs[END_EXPR];
  SourceLocation SingleLoc;

public:
  OmpSingleStmt(ASTContext &C, Stmt *Clause, Stmt *Body, SourceLocation SL);
  explicit OmpSingleStmt(EmptyShell Empty)
                          : Stmt(OmpSingleStmtClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    //FIXME?
    return SourceRange(SingleLoc, SubExprs[BODY]->getLocEnd());
  }

  child_range children() {
    //FIXME?
    return child_range(&SubExprs[0], &SubExprs[0]+END_EXPR);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpSingleStmtClass;
  }

  static bool classof(const OmpSingleStmt *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpTaskStmt - This class represents AST node for OpenMP directive 
// statement /{ "task" }/.
//===----------------------------------------------------------------------===//
class OmpTaskStmt : public Stmt {
  enum { CLAUSE, BODY, END_EXPR };
  Stmt* SubExprs[END_EXPR];
  SourceLocation TaskLoc;

public:
  OmpTaskStmt(ASTContext &C, Stmt *Clause, Stmt *Body, SourceLocation TL);
  explicit OmpTaskStmt(EmptyShell Empty)
                          : Stmt(OmpTaskStmtClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    //FIXME?
    return SourceRange(TaskLoc, SubExprs[BODY]->getLocEnd());
  }

  child_range children() {
    //FIXME?
    return child_range(&SubExprs[0], &SubExprs[0]+END_EXPR);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpTaskStmtClass;
  }

  static bool classof(const OmpTaskStmt *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpMasterStmt - This class represents AST node for OpenMP directive 
// statement /{ "master" }/.
//===----------------------------------------------------------------------===//
class OmpMasterStmt : public Stmt {
  Stmt* SubStmt;
  SourceLocation MasterLoc;

public:
  OmpMasterStmt(ASTContext &C, Stmt *Body, SourceLocation ML);
  explicit OmpMasterStmt(EmptyShell Empty)
                          : Stmt(OmpMasterStmtClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    //FIXME?
    return SourceRange(MasterLoc, SubStmt->getLocEnd());
  }

  child_range children() {
    //FIXME?
    return child_range(&SubStmt, &SubStmt+1);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpMasterStmtClass;
  }

  static bool classof(const OmpMasterStmt *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpBarrierStmt - This class represents AST node for OpenMP directive 
// statement /{ "barrier" }/.
//===----------------------------------------------------------------------===//
class OmpBarrierStmt : public Stmt {
  SourceLocation BarrierLoc;

public:
  OmpBarrierStmt(ASTContext &C, SourceLocation BL);
  explicit OmpBarrierStmt(EmptyShell Empty)
                          : Stmt(OmpBarrierStmtClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(BarrierLoc);
  }

  child_range children() { return child_range();
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpBarrierStmtClass;
  }

  static bool classof(const OmpBarrierStmt *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpTaskwaitStmt - This class represents AST node for OpenMP directive 
// statement /{ "taskwait" }/.
//===----------------------------------------------------------------------===//
class OmpTaskwaitStmt : public Stmt {
  SourceLocation TaskwaitLoc;

public:
  OmpTaskwaitStmt(ASTContext &C, SourceLocation TL);
  explicit OmpTaskwaitStmt(EmptyShell Empty)
                          : Stmt(OmpTaskwaitStmtClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(TaskwaitLoc);
  }

  child_range children() { return child_range();
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpTaskwaitStmtClass;
  }

  static bool classof(const OmpTaskwaitStmt *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpTaskyieldStmt - This class represents AST node for OpenMP directive 
// statement /{ "taskyield" }/.
//===----------------------------------------------------------------------===//
class OmpTaskyieldStmt : public Stmt {
  SourceLocation TaskyieldLoc;

public:
  OmpTaskyieldStmt(ASTContext &C, SourceLocation TL);
  explicit OmpTaskyieldStmt(EmptyShell Empty)
                          : Stmt(OmpTaskyieldStmtClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(TaskyieldLoc);
  }

  child_range children() { return child_range();
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpTaskyieldStmtClass;
  }

  static bool classof(const OmpTaskyieldStmt *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpOrderedStmt - This class represents AST node for OpenMP directive 
// statement /{ "ordered" }/.
//===----------------------------------------------------------------------===//
class OmpOrderedStmt : public Stmt {
  Stmt* SubStmt;
  SourceLocation OrderedLoc;

public:
  OmpOrderedStmt(ASTContext &C, Stmt *Body, SourceLocation OL);
  explicit OmpOrderedStmt(EmptyShell Empty)
                          : Stmt(OmpOrderedStmtClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    //FIXME?
    return SourceRange(OrderedLoc, SubStmt->getLocEnd());
  }

  child_range children() {
    //FIXME?
    return child_range(&SubStmt, &SubStmt+1);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpOrderedStmtClass;
  }

  static bool classof(const OmpOrderedStmt *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpThreadPrivateStmt - This class represents AST node for OpenMP pragma
// statement "threadprivate". 
//===----------------------------------------------------------------------===//
class OmpThreadPrivateStmt : public Stmt {
  Stmt** DeclRefExprs;
  SourceLocation ThreadPrivateLoc;
  SourceLocation LParenLoc;
  SourceLocation RParenLoc;

public:
  OmpThreadPrivateStmt(ASTContext &C, Expr **ExprStart, unsigned NumExprs,
        SourceLocation TPC, SourceLocation LP, SourceLocation RP);

  explicit OmpThreadPrivateStmt(EmptyShell Empty)
                          : Stmt(OmpThreadPrivateStmtClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(ThreadPrivateLoc, RParenLoc);
  }

  child_range children() {
    return child_range(&DeclRefExprs[0],
                       &DeclRefExprs[0]+OmpClauseBits.NumDeclRefExprs);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpThreadPrivateStmtClass;
  }

  static bool classof(const OmpThreadPrivateStmt *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpFlushStmt - This class represents AST node for OpenMP pragma statement
// "flush". 
//===----------------------------------------------------------------------===//
class OmpFlushStmt : public Stmt {
  Stmt** DeclRefExprs;
  SourceLocation FlushLoc;
  SourceLocation LParenLoc;
  SourceLocation RParenLoc;

public:
  OmpFlushStmt(ASTContext &C, Expr **ExprStart, unsigned NumExprs,
        SourceLocation FC, SourceLocation LP, SourceLocation RP);

  explicit OmpFlushStmt(EmptyShell Empty)
                          : Stmt(OmpFlushStmtClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    return SourceRange(FlushLoc, RParenLoc);
  }

  child_range children() {
    return child_range(&DeclRefExprs[0],
                       &DeclRefExprs[0]+OmpClauseBits.NumDeclRefExprs);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpFlushStmtClass;
  }

  static bool classof(const OmpFlushStmt *) { return true; }
};

//===----------------------------------------------------------------------===//
// OmpAtomicStmt - This class represents AST node for OpenMP directive 
// statement /{ "atomic" }/.
//===----------------------------------------------------------------------===//
class OmpAtomicStmt : public Stmt {
  enum { CLAUSE, BODY, END_EXPR };
  Stmt* SubExprs[END_EXPR];
  SourceLocation AtomicLoc;

public:
  OmpAtomicStmt(ASTContext &C, Stmt *Clause, Stmt *Body, SourceLocation AL,
                bool BodyIsExprStmt = false);
  explicit OmpAtomicStmt(EmptyShell Empty)
                          : Stmt(OmpAtomicStmtClass, Empty) { }

  SourceRange getSourceRange() const LLVM_READONLY {
    //FIXME?
    return SourceRange(AtomicLoc, SubExprs[BODY]->getLocEnd());
  }

  child_range children() {
    //FIXME?
    return child_range(&SubExprs[0], &SubExprs[0]+END_EXPR);
  }

  static bool classof(const Stmt *T) {
    return T->getStmtClass() == OmpAtomicStmtClass;
  }

  static bool classof(const OmpAtomicStmt *) { return true; }
};


}  // end namespace clang

#endif
